"""VPX platform."""
import asyncio
from typing import Dict, Optional

import logging

from mpf.platforms.interfaces.driver_platform_interface import DriverPlatformInterface, PulseSettings, HoldSettings
from mpf.platforms.interfaces.light_platform_interface import LightPlatformInterface
from mpf.platforms.interfaces.switch_platform_interface import SwitchPlatformInterface
from mpf.core.platform import LightsPlatform, SwitchPlatform, DriverPlatform, SwitchSettings, DriverSettings, \
    SwitchConfig, DriverConfig, RepulseSettings


class VirtualPinballSwitch(SwitchPlatformInterface):

    """A switch in VPX."""

    __slots__ = ["state"]

    def __init__(self, config, number):
        """Initialise switch."""
        super().__init__(config, number)
        self.state = self.config.invert

    def get_board_name(self):
        """Return the name of the board of this switch."""
        return "VPX"


class VirtualPinballLight(LightPlatformInterface):

    """A light in VPX."""

    __slots__ = ["_current_fade", "subtype", "hw_number", "machine"]

    def __init__(self, number, subtype, hw_number, machine):
        """Initialise LED."""
        super().__init__(number)
        self._current_fade = (0, -1, 0, -1)
        self.subtype = subtype
        self.hw_number = hw_number
        self.machine = machine

    @property
    def current_brightness(self) -> float:
        """Return current brightness."""
        current_time = self.machine.clock.get_time()
        start_brightness, start_time, target_brightness, target_time = self._current_fade
        if target_time > current_time:
            ratio = ((current_time - start_time) /
                     (target_time - start_time))
            return start_brightness + (target_brightness - start_brightness) * ratio

        return target_brightness

    def set_fade(self, start_brightness, start_time, target_brightness, target_time):
        """Set fade."""
        self._current_fade = (start_brightness, start_time, target_brightness, target_time)

    def get_board_name(self):
        """Return the name of the board of this light."""
        return "VPX"

    def is_successor_of(self, other):
        """Not implemented."""
        raise AssertionError("Not implemented. Let us know if you need it.")

    def get_successor_number(self):
        """Not implemented."""
        raise AssertionError("Not implemented. Let us know if you need it.")

    def __lt__(self, other):
        """Not implemented."""
        raise AssertionError("Not implemented. Let us know if you need it.")


class VirtualPinballDriver(DriverPlatformInterface):

    """A driver in VPX."""

    __slots__ = ["clock", "_state"]

    def __init__(self, config, number, clock):
        """Initialise virtual driver to disabled."""
        super().__init__(config, number)
        self.clock = clock
        self._state = False

    def get_board_name(self):
        """Return the name of the board of this driver."""
        return "VPX"

    def disable(self):
        """Disable virtual coil."""
        self._state = False

    def enable(self, pulse_settings: PulseSettings, hold_settings: HoldSettings):
        """Enable virtual coil."""
        del pulse_settings, hold_settings
        self._state = True

    def pulse(self, pulse_settings: PulseSettings):
        """Pulse virtual coil."""
        self._state = self.clock.get_time() + (pulse_settings.duration / 1000.0)

    @property
    def state(self) -> bool:
        """Return current state."""
        # pylint: disable-msg=no-else-return
        if isinstance(self._state, bool):
            return self._state
        else:
            return bool(self.clock.get_time() < self._state)


class VirtualPinballPlatform(LightsPlatform, SwitchPlatform, DriverPlatform):

    """VPX platform."""

    __slots__ = ["_lights", "_switches", "_drivers", "_last_drivers", "_last_lights", "_started", "rules"]

    def __init__(self, machine):
        """Initialise VPX platform."""
        super().__init__(machine)
        self._lights = {}           # type: Dict[str, VirtualPinballLight]
        self._switches = {}         # type: Dict[str, VirtualPinballSwitch]
        self._drivers = {}          # type: Dict[str, VirtualPinballDriver]
        self._last_drivers = {}     # type: Dict[str, bool]
        self._last_lights = {}      # type: Dict[str, bool]
        self._started = asyncio.Event()
        self.log = logging.getLogger("VPX Platform")
        self.log.debug("Configuring VPX hardware interface.")
        self.rules = {}

    async def initialize(self):
        """Initialise platform."""
        self.machine.bcp.interface.register_command_callback("vpcom_bridge", self._dispatch)
        self.machine.events.add_async_handler("init_phase_5", self._wait_for_connect)

    async def _wait_for_connect(self):
        """Wait until VPX connects."""
        await self._started.wait()

    async def _dispatch(self, client, subcommand=None, **kwargs):
        """Dispatch a VPX COM call."""
        self.log.debug("Got command %s args: %s", subcommand, kwargs)
        if not subcommand:
            self.machine.bcp.transport.send_to_client(client, "vpcom_bridge_response", error="command missing")
        try:
            method = getattr(self, "vpx_" + subcommand)
        except AttributeError:
            self.machine.bcp.transport.send_to_client(client, "vpcom_bridge_response",
                                                      error="Unknown command {}".format(subcommand))
            return

        try:
            result = method(**kwargs)
        # pylint: disable-msg=broad-except
        except Exception as e:
            self.machine.bcp.transport.send_to_client(client, "vpcom_bridge_response",
                                                      error="Exception: {}".format(e))
            return
        self.machine.bcp.transport.send_to_client(client, "vpcom_bridge_response", result=result)

    def vpx_start(self):
        """Start machine."""
        self._started.set()
        return True

    def vpx_get_switch(self, number):
        """Return switch value."""
        # pylint: disable-msg=no-else-return
        if self._switches[str(number)].config.invert:
            return not self._switches[str(number)].state
        else:
            return self._switches[str(number)].state

    def vpx_switch(self, number):
        """Return switch value."""
        return self.vpx_get_switch(number)

    def vpx_set_switch(self, number, value):
        """Update switch from VPX."""
        self._switches[str(number)].state = value
        self.machine.switch_controller.process_switch_by_num(state=1 if value else 0,
                                                             num=str(number),
                                                             platform=self)
        return True

    def vpx_pulsesw(self, number):
        """Pulse switch from VPX."""
        self._switches[str(number)].state = True
        self.machine.switch_controller.process_switch_by_num(state=1,
                                                             num=str(number),
                                                             platform=self)
        self._switches[str(number)].state = False
        self.machine.switch_controller.process_switch_by_num(state=0,
                                                             num=str(number),
                                                             platform=self)
        return True

    def vpx_changed_solenoids(self):
        """Return changed solenoids since last call."""
        changed_drivers = []
        for number, driver in self._drivers.items():
            if driver.state != self._last_drivers[number]:
                changed_drivers.append((number, driver.state))
                self._last_drivers[number] = driver.state

        return changed_drivers

    def _get_changed_lights_by_subtype(self, subtype):
        """Return changed lights since last call."""
        changed_lamps = []
        for number, light in self._lights.items():
            if light.subtype != subtype:
                continue
            brightness = light.current_brightness
            state = bool(brightness > 0.5)
            if state != self._last_lights[number]:
                changed_lamps.append((light.hw_number, state))
                self._last_lights[number] = state

        return changed_lamps

    def vpx_changed_lamps(self):
        """Return changed lamps since last call."""
        return self._get_changed_lights_by_subtype("matrix")

    def vpx_changed_gi_strings(self):
        """Return changed lamps since last call."""
        return self._get_changed_lights_by_subtype("gi")

    def vpx_changed_leds(self):
        """Return changed lamps since last call."""
        return self._get_changed_lights_by_subtype("led")

    def vpx_changed_flashers(self):
        """Return changed lamps since last call."""
        return self._get_changed_lights_by_subtype("flasher")

    def vpx_mech(self, number):
        """Not implemented."""
        self.log.warning("Command \"mech\" unimplemented: %s", number)
        return True

    def vpx_get_mech(self, number):
        """Not implemented."""
        self.log.warning("Command \"get_mech\" unimplemented: %s", number)
        return True

    def vpx_set_mech(self, number, value):
        """Not implemented."""
        self.log.warning("Command \"set_mech\" unimplemented: %s %s", number, value)
        return True

    def configure_switch(self, number: str, config: SwitchConfig, platform_config: dict) -> "SwitchPlatformInterface":
        """Configure VPX switch."""
        number = str(number)
        switch = VirtualPinballSwitch(config, number)
        self._switches[number] = switch
        return switch

    def configure_driver(self, config: DriverConfig, number: str, platform_settings: dict) -> "DriverPlatformInterface":
        """Configure VPX driver."""
        number = str(number)
        driver = VirtualPinballDriver(config, number, self.machine.clock)
        self._drivers[number] = driver
        self._last_drivers[number] = False
        return driver

    def vpx_get_hardwarerules(self):
        """Return hardware rules."""
        hardware_rules = []
        for rswitchandcoil, hold in self.rules.items():
            hardware_rules.append((rswitchandcoil[0].number, rswitchandcoil[1].number, hold))
        return hardware_rules

    def set_pulse_on_hit_and_enable_and_release_rule(self, enable_switch: SwitchSettings, coil: DriverSettings):
        """Pulse on hit and hold."""
        if (enable_switch.hw_switch, coil.hw_driver) in self.rules:
            raise AssertionError("Overwrote a rule without clearing it first {} <-> {}".format(
                enable_switch.hw_switch, coil.hw_driver))

        self.rules[(enable_switch.hw_switch, coil.hw_driver)] = True

    def set_pulse_on_hit_and_release_and_disable_rule(self, enable_switch: SwitchSettings, eos_switch: SwitchSettings,
                                                      coil: DriverSettings,
                                                      repulse_settings: Optional[RepulseSettings]):
        """Pulse on hit, disable on disable_switch hit."""
        del eos_switch
        if (enable_switch.hw_switch, coil.hw_driver) in self.rules:
            raise AssertionError("Overwrote a rule without clearing it first {} <-> {}".format(
                enable_switch.hw_switch, coil.hw_driver))

        # eos_switch is missing here intentionally
        self.rules[(enable_switch.hw_switch, coil.hw_driver)] = False

    def set_pulse_on_hit_and_enable_and_release_and_disable_rule(self, enable_switch: SwitchSettings,
                                                                 eos_switch: SwitchSettings, coil: DriverSettings,
                                                                 repulse_settings: Optional[RepulseSettings]):
        """Pulse on hit and hold, disable on disable_switch hit."""
        del eos_switch
        if (enable_switch.hw_switch, coil.hw_driver) in self.rules:
            raise AssertionError("Overwrote a rule without clearing it first {} <-> {}".format(
                enable_switch.hw_switch, coil.hw_driver))

        # eos_switch is missing here intentionally
        self.rules[(enable_switch.hw_switch, coil.hw_driver)] = True

    def set_pulse_on_hit_and_release_rule(self, enable_switch: SwitchSettings, coil: DriverSettings):
        """Pulse on hit and hold."""
        if (enable_switch.hw_switch, coil.hw_driver) in self.rules:
            raise AssertionError("Overwrote a rule without clearing it first {} <-> {}".format(
                enable_switch.hw_switch, coil.hw_driver))

        self.rules[(enable_switch.hw_switch, coil.hw_driver)] = True

    def set_pulse_on_hit_rule(self, enable_switch: SwitchSettings,
                              coil: DriverSettings):
        """Pulse on hit and release."""
        if (enable_switch.hw_switch, coil.hw_driver) in self.rules:
            raise AssertionError("Overwrote a rule without clearing it first {} <-> {}".format(
                enable_switch.hw_switch, coil.hw_driver))

        self.rules[(enable_switch.hw_switch, coil.hw_driver)] = False

    def clear_hw_rule(self, switch: SwitchSettings, coil: DriverSettings):
        """Clear hw rule."""
        if (switch.hw_switch, coil.hw_driver) in self.rules:
            del self.rules[(switch.hw_switch, coil.hw_driver)]
        else:
            self.log.debug("Tried to clear a non-existing rules %s <-> %s", switch, coil)

    def vpx_get_coilactive(self, number):
        """Return True if a MPF hw rule for the coil(number) exists."""
        for rswitchandcoil, _ in self.rules.items():
            if rswitchandcoil[1].number == number:
                return True

        return False

    async def get_hw_switch_states(self):
        """Return initial switch state."""
        hw_switches = {}
        for switch in self._switches.values():
            hw_switches[switch.number] = switch.state ^ switch.config.invert
        return hw_switches

    def configure_light(self, number: str, subtype: str, config, platform_settings: dict) -> "LightPlatformInterface":
        """Configure a VPX light."""
        del config
        if subtype and subtype not in ("gi", "matrix", "led", "flasher"):
            raise AssertionError("Unknown subtype: {}".format(subtype))
        if not subtype:
            subtype = "matrix"
        number = str(number)
        key = number + "-" + subtype
        light = VirtualPinballLight(key, subtype, number, self.machine)
        self._lights[key] = light
        self._last_lights[key] = False
        return light

    def parse_light_number_to_channels(self, number: str, subtype: str):
        """Parse channel str to a list of channels."""
        # pylint: disable-msg=no-else-return
        if subtype in ("gi", "matrix", "led", "flasher") or not subtype:
            return [
                {
                    "number": str(number)
                }
            ]
        else:
            raise AssertionError("Unknown subtype {}".format(subtype))



"""Test Randomizer class."""
from mpf.tests.MpfTestCase import MpfTestCase
from mpf.core.randomizer import Randomizer


class TestRandomizer(MpfTestCase):

    def get_config_file(self):
        return 'randomizer.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/randomizer/'

    def test_one_element_with_force_different(self):
        items = ['1']

        r = Randomizer(items)
        self.assertTrue(r.force_different)

        # it has one element and should thereby always return it
        self.assertEqual('1', next(r))
        self.assertEqual('1', next(r))
        self.assertEqual('1', next(r))

    def test_machine_randomizer(self):

        items = [
            '1',
            '2',
            '3',
        ]

        r = Randomizer(items)

        results = list()

        for x in range(10000):
            results.append(next(r))

        self.assertAlmostEqual(3333, results.count('1'), delta=500)
        self.assertAlmostEqual(3333, results.count('2'), delta=500)
        self.assertAlmostEqual(3333, results.count('3'), delta=500)

    def test_force_different(self):

        items = [
            ('1', 1),
            ('2', 1),
            ('3', 1),
        ]

        r = Randomizer(items)
        r.force_different = True

        last_item = None
        for x in range(1000):
            this_item = next(r)
            self.assertNotEqual(this_item, last_item)
            last_item = this_item

    def test_force_all(self):

        items = [
            ('1', 1),
            ('2', 1),
            ('3', 1),
        ]

        r = Randomizer(items)
        r.force_all = True

        last_item = None
        for x in range(100):
            results = set()
            results.add(next(r))
            self.assertNotEqual(last_item, r.get_current())
            results.add(next(r))
            results.add(next(r))
            last_item = r.get_current()
            self.assertEqual(len(results), 3)

    def test_no_loop(self):

        items = [
            ('1', 1),
            ('2', 1),
            ('3', 1),
        ]

        r = Randomizer(items)
        r.loop = False

        x = 0
        for _ in r:
            x += 1

        self.assertEqual(3, x)

    def test_weights(self):
        items = [
            ('1', 2),
            ('2', 1),
            ('3', 1),
        ]

        r = Randomizer(items)
        r.force_different = False

        results = list()

        for x in range(10000):
            results.append(next(r))

        self.assertAlmostEqual(5000, results.count('1'), delta=500)
        self.assertAlmostEqual(2500, results.count('2'), delta=500)
        self.assertAlmostEqual(2500, results.count('3'), delta=500)

        items = [
            ('1', 1),
            ('2', 1),
            ('3', 3),
        ]

        r = Randomizer(items)
        r.force_different = False

        results = list()

        for x in range(10000):
            results.append(next(r))

        self.assertAlmostEqual(2000, results.count('1'), delta=500)
        self.assertAlmostEqual(2000, results.count('2'), delta=500)
        self.assertAlmostEqual(6000, results.count('3'), delta=500)

        items = [
            ('1', 1),
            ('2', 6),
            ('3', 3),
        ]

        r = Randomizer(items)
        r.force_different = False

        results = list()

        for x in range(10000):
            results.append(next(r))

        self.assertAlmostEqual(1000, results.count('1'), delta=500)
        self.assertAlmostEqual(6000, results.count('2'), delta=500)
        self.assertAlmostEqual(3000, results.count('3'), delta=500)

    def test_loop_no_random(self):

        items = [
            ('1', 1),
            ('2', 1),
            ('3', 1),
        ]

        r = Randomizer(items)
        r.disable_random = True

        for i1 in range(50):
            self.assertEqual(next(r), '1')
            self.assertEqual(next(r), '2')
            self.assertEqual(next(r), '3')

    def test_no_loop_no_random(self):

        items = [
            ('1', 1),
            ('2', 1),
            ('3', 1),
        ]

        for _ in range(50):

            r = Randomizer(items)
            r.loop = False
            r.disable_random = True

            x = 0
            for i, result in enumerate(r):
                x += 1
                self.assertEqual(items[i][0], result)

            self.assertEqual(3, x)

    def test_fallback_value(self):

        items = []

        r = Randomizer(items)
        r.fallback_value = "foo"

        results = list()

        for x in range(100):
            results.append(next(r))

        self.assertEqual(100, results.count('foo'))


"""Command to build artifacts for non-dev operations."""
import pickle

import argparse
from mpf.core.utility_functions import Util

from mpf.core.config_loader import YamlMultifileConfigLoader, ProductionConfigLoader

from mpf.commands import MpfCommandLineParser

SUBCOMMAND = True


class Command(MpfCommandLineParser):

    """Build artifacts."""

    def __init__(self, args, path):
        """Parse args."""
        command_name = args.pop(1)
        super().__init__(args=args, path=path)

        machine_path, remaining_args = self.parse_args()
        self.machine_path = machine_path
        self.args = remaining_args

        parser = argparse.ArgumentParser(
            description='Build MPF production config.')

        parser.add_argument("-c",
                            action="store", dest="configfile",
                            default="config.yaml", metavar='config_file',
                            help="The name of a config file to load. Default "
                                 "is "
                                 "config.yaml. Multiple files can be used "
                                 "via a comma-"
                                 "separated list (no spaces between)")

        self.args = parser.parse_args(remaining_args)
        self.args.configfile = Util.string_to_event_list(self.args.configfile)

        method = getattr(self, command_name)
        method()

    def production_bundle(self):
        """Create a production bundle."""
        config_loader = YamlMultifileConfigLoader(self.machine_path, self.args.configfile, False, False)
        mpf_config = config_loader.load_mpf_config()
        mc_config = config_loader.load_mc_config()

        pickle.dump(mpf_config, open(ProductionConfigLoader.get_mpf_bundle_path(self.machine_path), "wb"))
        pickle.dump(mc_config, open(ProductionConfigLoader.get_mpf_mc_bundle_path(self.machine_path), "wb"))
        print("Success.")



"""Controller for all service functionality.

Controller provides all service information and can perform service tasks. Displaying the information is performed by
the service mode or other components.
"""
import re
from collections import namedtuple

from typing import List

from mpf.core.mpf_controller import MpfController

SwitchMap = namedtuple("SwitchMap", ["board", "switch"])
CoilMap = namedtuple("CoilMap", ["board", "coil"])
LightMap = namedtuple("LightMap", ["board", "light"])


class ServiceController(MpfController):

    """Provides all service information and can perform service tasks."""

    __slots__ = ["_enabled"]

    config_name = "service_controller"

    def __init__(self, machine):
        """Initialise service controller."""
        super().__init__(machine)
        self._enabled = False
        self.configure_logging("service")

    @staticmethod
    def _natural_key_sort(string_to_sort):
        """Sort by natural keys like humans do.

        See http://www.codinghorror.com/blog/archives/001018.html.
        """
        return [int(s) if s.isdigit() else s for s in re.split(r'(\d+)', string_to_sort)]

    def is_in_service(self) -> bool:
        """Return true if in service mode."""
        return self._enabled

    def start_service(self):
        """Start service mode."""
        if self.is_in_service():
            raise AssertionError("Already in service mode!")
        self._enabled = True

        self.info_log("Entered service mode. Resetting game if running. Resetting hardware interface now.")
        # this will stop attact and game mode
        for mode in self.machine.modes.values():
            if not mode.active or mode.name in ["service", "game"]:
                continue
            mode.stop()

        # explicitly stop game last
        if self.machine.modes["game"].active:
            self.machine.modes["game"].stop()

        self.machine.events.post("service_mode_entered")

    async def stop_service(self):
        """Stop service mode."""
        if not self.is_in_service():
            raise AssertionError("Not in service mode!")
        self._enabled = False

        # this event starts attract mode again
        self.machine.events.post("service_mode_exited")
        await self.machine.reset()

    # pylint: disable-msg=no-self-use
    def add_technical_alert(self, device, issue):
        """Add an alert about a technical problem."""
        del device
        del issue
        # this is prepared but not yet implemented in service mode

    def get_switch_map(self):
        """Return a map of all switches in the machine."""
        switch_map = []
        for switch in self.machine.switches.values():
            switch_map.append(SwitchMap(switch.hw_switch.get_board_name(), switch))

        # sort by board + driver number
        switch_map.sort(key=lambda x: (self._natural_key_sort(x[0]),
                                       self._natural_key_sort(str(x[1].hw_switch.number))))
        return switch_map

    def get_coil_map(self) -> List[CoilMap]:
        """Return a map of all coils in the machine."""
        coil_map = []
        for coil in self.machine.coils.values():
            assert coil.hw_driver is not None
            coil_map.append(CoilMap(coil.hw_driver.get_board_name(), coil))

        # sort by board + driver number
        coil_map.sort(key=lambda x: (self._natural_key_sort(x[0]), self._natural_key_sort(str(x[1].hw_driver.number))))
        return coil_map

    def get_light_map(self) -> List[LightMap]:
        """Return a map of all lights in the machine."""
        light_map = []
        for light in self.machine.lights.values():
            light_map.append(LightMap(next(iter(light.hw_drivers.values()))[0].get_board_name(), light))

        # sort by board + driver number
        light_map.sort(key=lambda x: (self._natural_key_sort(x[0]), self._natural_key_sort(str(x[1].config['number']))))
        return light_map



"""Baseclass for ball device ball counters.

The duty of this device is to maintain the current ball count of the device.
"""
import asyncio

from typing import List, Optional
from mpf.core.utility_functions import Util

MYPY = False
if MYPY:    # pragma: no cover
    from mpf.devices.ball_device.ball_device import BallDevice  # pylint: disable-msg=cyclic-import,unused-import
    from mpf.core.machine import MachineController  # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.ball_device.ball_count_handler import BallCountHandler     # pylint: disable-msg=cyclic-import,unused-import; # noqa


class EjectTracker:

    """Tracks ball changes during an eject."""

    __slots__ = ["machine", "_already_left", "_ball_count_handler", "_task", "_event_queue", "_ball_left",
                 "_ball_returned", "_ready", "_unknown_balls", "_num_unknown_balls", "_num_lost_balls"]

    def __init__(self, ball_counter_handler, already_left):
        """Initialise eject tracker."""
        self.machine = ball_counter_handler.machine
        self._already_left = already_left
        self._ball_count_handler = ball_counter_handler     # type: BallCountHandler
        self._task = None
        self._event_queue = asyncio.Queue()
        self._ball_left = asyncio.Future()
        self._ball_returned = asyncio.Future()
        self._ready = asyncio.Future()
        self._unknown_balls = asyncio.Future()
        self._num_unknown_balls = 0
        self._num_lost_balls = 0

    async def will_eject(self):
        """Start process."""
        await self._ball_count_handler.counter.wait_for_count_stable()
        ball_changes = self._ball_count_handler.counter.register_change_stream()
        if not self._already_left:
            ball_left = await self._ball_count_handler.counter.wait_for_ball_to_leave()
            self._ball_left = asyncio.ensure_future(ball_left)

        self._task = self.machine.clock.loop.create_task(self._run(ball_changes))
        self._task.add_done_callback(Util.raise_exceptions)

    async def _run(self, ball_changes):
        already_left = self._already_left
        while True:
            change = await ball_changes.get()
            if isinstance(change, BallLostActivity) and not already_left and self._ball_left.done():
                already_left = True
                self._ball_count_handler.ball_device.debug_log("Got ball left during eject")
                continue

            if isinstance(change, BallLostActivity):
                self.track_lost_balls(1)
            elif isinstance(change, BallEntranceActivity):
                await self.track_ball_entrance()
            elif isinstance(change, UnknownBallActivity):
                self.track_unknown_balls(1)
            elif isinstance(change, BallReturnActivity):
                self.track_ball_returned()
            else:
                raise AssertionError("Unknown activity {}".format(change))

    def cancel(self):
        """Cancel eject tracker."""
        if self._task:
            self._task.cancel()
        if not self._ball_left.done():
            self._ball_left.cancel()

    def is_jammed(self):
        """Return true if currently jammed."""
        return self._ball_count_handler.counter.is_jammed()

    def track_ball_returned(self):
        """Track ball returned."""
        self._ball_count_handler.ball_device.debug_log("Got ball return during eject")
        self._ball_returned.set_result(True)

    async def track_ball_entrance(self):
        """Track ball entrance."""
        self._ball_count_handler.ball_device.debug_log("Got ball entrance during eject")
        await self._ball_count_handler.entrance_during_eject()

    def track_unknown_balls(self, balls):
        """Track unknown ball."""
        self._ball_count_handler.ball_device.debug_log("Got %s unknown ball during eject", balls)
        self._num_unknown_balls += balls
        if not self._unknown_balls.done():
            self._unknown_balls.set_result(True)

    def get_num_unknown_balls(self):
        """Return unknown balls."""
        return self._num_unknown_balls

    def track_lost_balls(self, balls):
        """Track lost ball."""
        self._num_lost_balls += balls
        if self._num_lost_balls >= self._num_unknown_balls and self._unknown_balls.done():
            self._unknown_balls = asyncio.Future()

    def wait_for_ball_return(self):
        """Wait until a ball returned."""
        return asyncio.shield(self._ball_returned)

    def wait_for_ball_unknown_ball(self):
        """Return true if the device has unknown balls which are neither clearly new or returned."""
        return asyncio.shield(self._unknown_balls)

    def wait_for_ball_left(self):
        """Wait until a ball left."""
        if self._already_left:
            raise AssertionError("Invalid wait. Ball left before eject.")
        return asyncio.shield(self._ball_left)

    def wait_for_ready(self):
        """Wait until the device is ready."""
        return asyncio.shield(self._ready)

    def set_ready(self):
        """Set device ready."""
        self._ready.set_result("ready")


class BallActivity:

    """An acticity in a ball device."""

    __slots__ = []  # type: List[str]


class BallLostActivity(BallActivity):

    """A ball was lost/ejected from device."""

    __slots__ = []  # type: List[str]


class NewBallActivity(BallActivity):

    """A new ball was found in the device."""

    __slots__ = []  # type: List[str]


class BallEntranceActivity(NewBallActivity):

    """A new ball entered the device (did not return)."""

    __slots__ = []  # type: List[str]


class BallReturnActivity(NewBallActivity):

    """A ball returned."""

    __slots__ = []  # type: List[str]


class UnknownBallActivity(NewBallActivity):

    """A unknown new ball was found in the device.

    This could be a returned or entered ball.
    """

    __slots__ = []  # type: List[str]


class PhysicalBallCounter:

    """Ball counter for ball device."""

    __slots__ = ["ball_device", "config", "machine", "_last_count", "_count_stable", "_activity_queues",
                 "_ball_change_futures"]

    def __init__(self, ball_device, config) -> None:
        """Initialise ball counter."""
        self.ball_device = ball_device              # type: BallDevice
        self.config = config
        self.machine = self.ball_device.machine     # type: MachineController

        self._last_count = None                     # type: Optional[int]
        self._count_stable = asyncio.Event()
        self._activity_queues = []                  # type: List[asyncio.Queue[BallActivity]]
        self._ball_change_futures = []              # type: List[asyncio.Future]

    def stop(self):
        """Stop counter."""

    def invalidate_count(self):
        """Invalidate the count."""
        self._count_stable.clear()

    def mark_count_as_stable_and_trigger_activity(self):
        """Mark count as stable and trigger activity."""
        self._count_stable.set()
        self.trigger_activity()

    def trigger_activity(self):
        """Trigger all activity futures."""
        for future in self._ball_change_futures:
            if not future.done():
                future.set_result(True)
        self._ball_change_futures = []

    def debug_log(self, *args, **kwargs):
        """Debug log."""
        self.ball_device.debug_log(*args, **kwargs)

    def count_balls_sync(self) -> int:
        """Return the number of current active switches or raises ValueError when count is not stable."""
        raise NotImplementedError()

    def is_jammed(self) -> bool:
        """Return true if device is jammed."""
        raise NotImplementedError()

    def is_count_unreliable(self) -> bool:
        """Return true if device is jammed and cannot count."""
        raise NotImplementedError()

    async def count_balls(self) -> int:
        """Return the current ball count."""
        # wait until count is stable
        await self._count_stable.wait()
        assert self._last_count is not None
        return self._last_count

    def wait_for_count_stable(self):
        """Wait for stable count."""
        return asyncio.ensure_future(self._count_stable.wait())

    @property
    def is_ready_to_receive(self):
        """Return true if the counter is ready to receive."""
        raise NotImplementedError()

    async def wait_for_ready_to_receive(self):
        """Wait until the counter is ready to count an incoming ball."""
        raise NotImplementedError()

    async def wait_for_ball_to_leave(self):
        """Wait until a ball left."""
        raise NotImplementedError()

    def received_entrance_event(self):
        """Handle entrance event."""
        raise NotImplementedError()

    def register_change_stream(self):
        """Register queue which returns all changes."""
        queue = asyncio.Queue()
        self._activity_queues.append(queue)
        return queue

    def record_activity(self, activity_type: BallActivity) -> None:
        """Record an activity."""
        for queue in self._activity_queues:
            queue.put_nowait(activity_type)

    def wait_for_ball_activity(self):
        """Wait for (settled) ball activity in device."""
        future = asyncio.Future()
        self._ball_change_futures.append(future)
        return future

    @property
    def capacity(self):
        """Return capacity under normal circumstances.

        This should not include jam switches or similar overflow mechanisms.
        """
        raise NotImplementedError()

    # pylint: disable-msg=inconsistent-return-statements
    async def wait_for_ball_count_changes(self, old_count: int):
        """Wait for ball count changes and return the new count.

        Args:
        ----
            old_count: Old ball count. Will return when the current count differs
        """
        while True:
            current_count = await self.count_balls()
            if current_count != old_count:
                return current_count

            await self.wait_for_ball_activity()



"""Contains the Light class."""
import asyncio

from functools import partial

from typing import Set, Dict, List, Tuple, Any

from mpf.core.delays import DelayManager

from mpf.core.platform import LightsPlatform, LightConfig, LightConfigColors

from mpf.core.device_monitor import DeviceMonitor
from mpf.core.machine import MachineController
from mpf.core.rgb_color import RGBColor, ColorException
from mpf.core.system_wide_device import SystemWideDevice
from mpf.devices.device_mixins import DevicePositionMixin

MYPY = False
if MYPY:
    from mpf.platforms.interfaces.light_platform_interface import LightPlatformInterface    # pylint: disable-msg=cyclic-import,unused-import; # noqa


class LightStackEntry:

    """Data class for a light stack entry."""

    __slots__ = ["priority", "start_time", "start_color", "dest_time", "dest_color", "key"]

    # pylint: disable-msg=too-many-arguments
    def __init__(self, priority, key, start_time, start_color, dest_time, dest_color):
        """Initialize light stack entry."""
        self.priority = priority
        self.key = key
        self.start_time = start_time
        self.start_color = start_color
        self.dest_time = dest_time
        self.dest_color = dest_color

    def __gt__(self, other):
        """Compare two stack entries."""
        return self.priority > other.priority or (self.priority == other.priority and self.key > other.key)

    def __repr__(self):
        """Return string representation."""
        return "<LightStackEntry {}: {} ({}) -> {} ({}) Prio: {}>".format(
            self.key, self.start_color, self.start_time, self.dest_color, self.dest_time, self.priority)


@DeviceMonitor(_color="color", _do_not_overwrite_setter=True)
class Light(SystemWideDevice, DevicePositionMixin):

    """A light in a pinball machine."""

    config_section = 'lights'
    collection = 'lights'
    class_label = 'light'

    __slots__ = ["hw_drivers", "platforms", "delay", "default_fade_ms", "_color_correction_profile", "stack",
                 "_off_color", "_drivers_loaded", "_last_fade_target"]

    def __init__(self, machine, name):
        """Initialise light."""
        self.hw_drivers = {}        # type: Dict[str, List[LightPlatformInterface]]
        self.platforms = set()      # type: Set[LightsPlatform]
        super().__init__(machine, name)
        self.machine.light_controller.initialise_light_subsystem()
        self.delay = DelayManager(self.machine)
        self._drivers_loaded = asyncio.Future()

        self.default_fade_ms = None
        self._off_color = RGBColor("off")

        self._color_correction_profile = None
        self._last_fade_target = None

        self.stack = list()     # type: List[LightStackEntry]
        """A list of dicts which represents different commands that have come
        in to set this light to a certain color (and/or fade). Each entry in the
        list contains the following key/value pairs:

        priority:
            The relative priority of this color command. Higher numbers
            take precedent, and the highest priority entry will be the command
            that's currently active. In the event of a tie, whichever entry was
            added last wins (based on 'start_time' below).
        start_time:
            The clock time when this command was added. Primarily used
            to calculate fades, but also used as a tie-breaker for multiple
            entries with the same priority.
        start_color:
            RGBColor() of the color of this light when this command came in.
        dest_time:
            Clock time that represents when a fade (from start_color to
            dest_color) will be done. If this is 0, that means there is no
            fade. When a fade is complete, this value is reset to 0.
        dest_color:
            RGBColor() of the destination this light is fading to. If
            a command comes in with no fade, then this will be the same as the
            'color' below.
        key:
            An arbitrary unique identifier to keep multiple entries in the
            stack separate. If a new color command comes in with a key that
            already exists for an entry in the stack, that entry will be
            replaced by the new entry. The key is also used to remove entries
            from the stack (e.g. when shows or modes end and they want to
            remove their commands from the light).
        """

    @classmethod
    def device_class_init(cls, machine: MachineController):
        """Register handler for duplicate light number checks."""
        machine.events.add_handler("init_phase_4",
                                   cls._check_duplicate_light_numbers,
                                   machine=machine)

    def get_hw_numbers(self):
        """Return a list of all hardware driver numbers."""
        numbers = []
        for _, drivers in sorted(self.hw_drivers.items()):
            for driver in sorted(drivers, key=lambda x: x.number):
                numbers.append(driver.number)

        return numbers

    @staticmethod
    def _check_duplicate_light_numbers(machine: MachineController, **kwargs):
        del kwargs
        check_set = set()
        for light in machine.lights.values():
            for drivers in light.hw_drivers.values():
                for driver in drivers:
                    key = (light.config['platform'], driver.number, type(driver))
                    if key in check_set:
                        raise AssertionError(
                            "Duplicate light number {} {} for light {}".format(
                                type(driver), driver.number, light))

                    check_set.add(key)

    def _map_channels_to_colors(self, channel_list) -> dict:
        if self.config['type']:
            color_channels = self.config['type']
        else:
            if len(channel_list) == 1:
                # for one channel default to a white channel
                color_channels = "w"
            elif len(channel_list) == 3:
                # for three channels default to RGB
                color_channels = "rgb"
            else:
                raise AssertionError("Please provide a type for light {}. No default for channels {}.".
                                     format(self.name, channel_list))

        if len(channel_list) != len(color_channels):
            raise AssertionError("Type {} does not match channels {} for light {}".format(
                color_channels, channel_list, self.name
            ))

        channels = {}   # type: Dict[str, List[Any]]
        for color_name in color_channels:
            # red channel
            if color_name == 'r':
                full_color_name = "red"
            # green channel
            elif color_name == 'g':
                full_color_name = "green"
            # blue channel
            elif color_name == 'b':
                full_color_name = "blue"
            # simple white channel
            elif color_name == 'w':
                full_color_name = "white"
            else:
                raise AssertionError("Invalid element {} in type {} of light {}".format(
                    color_name, self.config['type'], self.name))

            if full_color_name not in channels:
                channels[full_color_name] = []
            channels[full_color_name].append(channel_list.pop(0))

        return channels

    def wait_for_loaded(self):
        """Return future."""
        return asyncio.shield(self._drivers_loaded)

    def get_successor_number(self):
        """Get the number of the next light channel.

        We first have to find the last channel and then get the next number based on that.
        """
        all_drivers = []
        for drivers in self.hw_drivers.values():
            all_drivers.extend(drivers)
        sorted_channels = sorted(all_drivers)
        return sorted_channels[-1].get_successor_number()

    def _load_hw_driver_sequentially(self, next_channel):
        if self.config['number'] or self.config['channels']:
            self.raise_config_error("Cannot use start_channel/previous and number or channels.", 3)
        if not self.config['type']:
            self.raise_config_error("Cannot use previous or start_channel without type. "
                                    "Add a type setting to your light.", 2)

        for color_name in self.config['type']:
            # red channel
            if color_name == 'r':
                full_color_name = "red"
            # green channel
            elif color_name == 'g':
                full_color_name = "green"
            # blue channel
            elif color_name == 'b':
                full_color_name = "blue"
            # simple white channel
            elif color_name == 'w':
                full_color_name = "white"
            else:
                raise AssertionError("Invalid element {} in type {} of light {}".format(
                    color_name, self.config['type'], self.name))

            if full_color_name not in self.hw_drivers:
                self.hw_drivers[full_color_name] = []
            channel = {'subtype': self.config['subtype'], 'platform': self.config['platform'],
                       'platform_settings': self.config['platform_settings'], 'number': next_channel}
            channel = self.machine.config_validator.validate_config("light_channels", channel)
            driver = self._load_hw_driver(channel, full_color_name)
            next_channel = driver.get_successor_number()
            self.hw_drivers[full_color_name].append(driver)

    def _load_hw_drivers(self):
        if not self.config['channels']:
            # get channels from number + platform
            platform = self.machine.get_platform_sections('lights', self.config['platform'])
            platform.assert_has_feature("lights")
            try:
                channel_list = platform.parse_light_number_to_channels(self.config['number'], self.config['subtype'])
            except AssertionError as e:
                raise AssertionError("Failed to parse light number {} in platform. See error above".
                                     format(self.name)) from e

            # copy platform and platform_settings to all channels
            for channel, _ in enumerate(channel_list):
                channel_list[channel]['subtype'] = self.config['subtype']
                channel_list[channel]['platform'] = self.config['platform']
                channel_list[channel]['platform_settings'] = self.config['platform_settings']
            # map channels to colors
            channels = self._map_channels_to_colors(channel_list)
        else:
            if self.config['number'] or self.config['platform'] or self.config['platform_settings']:
                raise AssertionError("Light {} cannot contain platform/platform_settings/number and channels".
                                     format(self.name))
            # alternatively use channels from config
            channels = self.config['channels']
            # ensure that we got lists
            for channel in channels:
                if not isinstance(channels[channel], list):
                    channels[channel] = [channels[channel]]

        if not channels:
            raise AssertionError("Light {} has no channels.".format(self.name))

        for color, channel_list in channels.items():
            self.hw_drivers[color] = []
            for channel in channel_list:
                channel = self.machine.config_validator.validate_config("light_channels", channel)
                driver = self._load_hw_driver(channel, color)
                self.hw_drivers[color].append(driver)

    def _load_hw_driver(self, channel, color):
        """Load one channel."""
        platform = self.machine.get_platform_sections('lights', channel['platform'])
        self.platforms.add(platform)

        if not platform.features['allow_empty_numbers'] and channel['number'] is None:
            self.raise_config_error("Light must have a number.", 1)

        config = LightConfig(
            name=self.name,
            color=LightConfigColors[color.upper()]
        )

        try:
            return platform.configure_light(channel['number'], channel['subtype'], config, channel['platform_settings'])
        except AssertionError as e:
            raise AssertionError("Failed to configure light {} in platform. See error above".
                                 format(self.name)) from e

    async def _initialize(self):
        await super()._initialize()
        if self.config['previous']:
            await self.config['previous'].wait_for_loaded()
            start_channel = self.config['previous'].get_successor_number()
            self._load_hw_driver_sequentially(start_channel)
        elif self.config['start_channel']:
            self._load_hw_driver_sequentially(self.config['start_channel'])
        else:
            self._load_hw_drivers()
        self._drivers_loaded.set_result(True)

        self.config['default_on_color'] = RGBColor(self.config['default_on_color'])

        if self.config['color_correction_profile'] is not None:
            profile_name = self.config['color_correction_profile']
        elif 'light_settings' in self.machine.config and \
                self.machine.config['light_settings']['default_color_correction_profile'] is not None:
            profile_name = self.machine.config['light_settings']['default_color_correction_profile']
        else:
            profile_name = None

        if profile_name:
            if profile_name in self.machine.light_controller.light_color_correction_profiles:
                profile = self.machine.light_controller.light_color_correction_profiles[profile_name]

                if profile is not None:
                    self._set_color_correction_profile(profile)
            else:   # pragma: no cover
                error = "Color correction profile '{}' was specified for light '{}'"\
                        " but the color correction profile does not exist."\
                        .format(profile_name, self.name)
                self.error_log(error)
                raise ValueError(error)

        if self.config['fade_ms'] is not None:
            self.default_fade_ms = self.config['fade_ms']
        else:
            self.default_fade_ms = (self.machine.config['light_settings']
                                    ['default_fade_ms'])

        self.debug_log("Initializing Light. CC Profile: %s, "
                       "Default fade: %sms", self._color_correction_profile,
                       self.default_fade_ms)

    def _set_color_correction_profile(self, profile):
        """Apply a color correction profile to this light.

        Args:
        ----
            profile: An RGBColorCorrectionProfile() instance

        """
        self._color_correction_profile = profile

    # pylint: disable-msg=too-many-arguments
    def color(self, color, fade_ms=None, priority=0, key=None, start_time=None):
        """Add or update a color entry in this light's stack.

        Calling this methods is how you tell this light what color you want it to be.

        Args:
        ----
            color: RGBColor() instance, or a string color name, hex value, or
                3-integer list/tuple of colors.
            fade_ms: Int of the number of ms you want this light to fade to the
                color in. A value of 0 means it's instant. A value of None (the
                default) means that it will use this light's and/or the machine's
                default fade_ms setting.
            priority: Int value of the priority of these incoming settings. If
                this light has current settings in the stack at a higher
                priority, the settings you're adding here won't take effect.
                However they're still added to the stack, so if the higher
                priority settings are removed, then the next-highest apply.
            key: An arbitrary identifier (can be any immutable object) that's
                used to identify these settings for later removal. If any
                settings in the stack already have this key, those settings
                will be replaced with these new settings.
            start_time: Time this occured to synchronize lights.
        """
        if self._debug:
            self.debug_log("Received color() command. color: %s, fade_ms: %s "
                           "priority: %s, key: %s", color, fade_ms, priority,
                           key)

        if isinstance(color, str) and color == "on":
            color = self.config['default_on_color']
        elif not isinstance(color, RGBColor):
            color = RGBColor(color)

        if fade_ms is None:
            fade_ms = self.default_fade_ms

        if not start_time:
            start_time = self.machine.clock.get_time()

        color_changes = not self.stack or self.stack[0].priority <= priority or self.stack[0].dest_color is None

        self._add_to_stack(color, fade_ms, priority, key, start_time)

        if color_changes:
            self._schedule_update()

    def on(self, brightness=None, fade_ms=None, priority=0, key=None, **kwargs):
        """Turn light on.

        Args:
        ----
            brightness: Brightness factor for "on".
            key: key for removal later on
            priority: priority on stack
            fade_ms: duration of fade
        """
        del kwargs
        color = self.config['default_on_color']
        if brightness is not None:
            color *= brightness / 255
        self.color(color=color, fade_ms=fade_ms,
                   priority=priority, key=key)

    def off(self, fade_ms=None, priority=0, key=None, **kwargs):
        """Turn light off.

        Args:
        ----
            key: key for removal later on
            priority: priority on stack
            fade_ms: duration of fade
        """
        del kwargs
        self.color(color=self._off_color, fade_ms=fade_ms, priority=priority,
                   key=key)

    # pylint: disable-msg=too-many-arguments
    def _add_to_stack(self, color, fade_ms, priority, key, start_time):
        """Add color to stack."""
        # handle None to make keys sortable
        if key is None:
            key = ""
        elif not isinstance(key, str):
            raise AssertionError("Key should be string")

        if self.stack and priority < self._get_priority_from_key(key):
            if self._debug:
                self.debug_log("Incoming priority %s is lower than an existing "
                               "stack item with the same key %s. Not adding to "
                               "stack.", priority, key)
            return

        if self.stack and self._debug and priority == self.stack[0].priority and key != self.stack[0].key:
            self.debug_log("Light stack contains two entries with the same priority %s but different keys: %s",
                           priority, self.stack)

        if fade_ms:
            dest_time = start_time + (fade_ms / 1000)
            color_below = self.get_color_below(priority, key)
        else:
            dest_time = 0
            color_below = None

        if self.stack:
            self._remove_from_stack_by_key(key)

        self.stack.append(LightStackEntry(priority,
                                          key,
                                          start_time,
                                          color_below,
                                          dest_time,
                                          color))

        if len(self.stack) > 1:
            self.stack.sort(reverse=True)

        if self._debug:
            self.debug_log("+-------------- Adding to stack ----------------+")
            self.debug_log("priority: %s", priority)
            self.debug_log("start_time: %s", self.machine.clock.get_time())
            self.debug_log("start_color: %s", color_below)
            self.debug_log("dest_time: %s", dest_time)
            self.debug_log("dest_color: %s", color)
            self.debug_log("key: %s", key)

    def remove_from_stack_by_key(self, key, fade_ms=None):
        """Remove a group of color settings from the stack.

        Args:
        ----
            key: The key of the settings to remove (based on the 'key'
                parameter that was originally passed to the color() method.)
            fade_ms: Time to fade out the light.

        This method triggers a light update, so if the highest priority settings
        were removed, the light will be updated with whatever's below it. If no
        settings remain after these are removed, the light will turn off.
        """
        if not self.stack:
            # no stack
            return

        if fade_ms is None:
            fade_ms = self.default_fade_ms

        key = str(key)

        priority = None
        color_changes = True
        stack = []
        for i, entry in enumerate(self.stack):
            if entry.key == key:
                stack = self.stack[i:]
                priority = entry.priority
                break
            if entry.dest_color is not None:
                # no transparency above key
                color_changes = False

        # key not in stack
        if not stack:
            return

        # this is already a fadeout. do not fade out the fade out.
        if stack[0].dest_color is None:
            fade_ms = None

        if fade_ms:
            # fade to underlaying color
            color_of_key = self._get_color_and_fade(stack, 0)[0]

            self._remove_from_stack_by_key(key)

            start_time = self.machine.clock.get_time()
            self.stack.append(LightStackEntry(priority,
                                              key,
                                              start_time,
                                              color_of_key,
                                              start_time + fade_ms / 1000.0,
                                              None))
            self.delay.reset(ms=fade_ms, callback=partial(self._remove_fade_out, key=key),
                             name="remove_fade_{}".format(key))
            if len(self.stack) > 1:
                self.stack.sort(reverse=True)
        else:
            # no fade -> just remove color from stack
            self._remove_from_stack_by_key(key)

        if color_changes:
            self._schedule_update()

    def _remove_fade_out(self, key):
        """Remove a timed out fade out."""
        if not self.stack:
            return

        found = False
        color_change = True
        for _, entry in enumerate(self.stack):
            if entry.key == key and entry.dest_color is None:
                found = True
                break
            if entry.dest_color is not None:
                # found entry above the removed which is non-transparent
                color_change = False

        if found:
            if self._debug:
                self.debug_log("Removing fadeout for key '%s' from stack", key)
            self.stack = [x for x in self.stack if x.key != key or x.dest_color is not None]

        if found and color_change:
            self._schedule_update()

    def _remove_from_stack_by_key(self, key):
        """Remove a key from stack."""
        # tune the common case
        if not self.stack:
            return
        if self._debug:
            self.debug_log("Removing key '%s' from stack", key)
        if len(self.stack) == 1:
            if self.stack[0].key == key:
                self.stack = []
        else:
            self.stack = [x for x in self.stack if x.key != key]

    def _schedule_update(self):
        start_color, start_time, target_color, target_time = self._get_color_and_target_time(self.stack)

        # check if our fade target really changed
        if (start_color, start_time, target_color, target_time) == self._last_fade_target:
            # nope its the same -> nothing to do
            return

        if self._last_fade_target and target_color == self._last_fade_target[2] and \
                (self._last_fade_target[3] < 0 or self._last_fade_target[3] < self.machine.clock.get_time()):
            # last fade had the same target and finished already -> nothing to do
            return

        self._last_fade_target = (start_color, start_time, target_color, target_time)

        if start_color != target_color:
            start_color = self.color_correct(self.gamma_correct(start_color))
            target_color = self.color_correct(self.gamma_correct(target_color))
        else:
            start_color = self.color_correct(self.gamma_correct(start_color))
            target_color = start_color

        for color, drivers in self.hw_drivers.items():
            if color in ["red", "blue", "green"]:
                start_brightness = getattr(start_color, color) / 255.0
                target_brightness = getattr(target_color, color) / 255.0
            elif color == "white":
                start_brightness = min(start_color.red, start_color.green, start_color.blue) / 255.0
                target_brightness = min(target_color.red, target_color.green, target_color.blue) / 255.0
            else:
                raise ColorException("Invalid color {}".format(color))
            for driver in drivers:
                driver.set_fade(start_brightness, start_time, target_brightness, target_time)

        for platform in self.platforms:
            platform.light_sync()

    def clear_stack(self):
        """Remove all entries from the stack and resets this light to 'off'."""
        self.stack = []

        if self._debug:
            self.debug_log("Clearing Stack")

        self._schedule_update()

    def _get_priority_from_key(self, key):
        if not self.stack:
            return 0
        if self.stack[0].key == key:
            return self.stack[0].priority
        try:
            return [x for x in self.stack if x.key == key][0].priority
        except IndexError:
            return 0

    def gamma_correct(self, color):
        """Apply max brightness correction to color.

        Args:
        ----
            color: The RGBColor() instance you want to have gamma applied.

        Returns an updated RGBColor() instance with gamma corrected.
        """
        factor = self.machine.light_controller.brightness_factor
        if factor == 1.0:
            return color

        return RGBColor([int(x * factor) for x in color])

    def color_correct(self, color):
        """Apply the current color correction profile to the color passed.

        Args:
        ----
            color: The RGBColor() instance you want to get color corrected.

        Returns an updated RGBColor() instance with the current color
        correction profile applied.

        Note that if there is no current color correction profile applied, the
        returned color will be the same as the color that was passed.
        """
        if self._color_correction_profile is None:
            return color

        if self._debug:
            self.debug_log("Applying color correction: %s (applied "
                           "'%s' color correction profile)",
                           self._color_correction_profile.apply(color),
                           self._color_correction_profile.name)

        return self._color_correction_profile.apply(color)

    def _get_color_and_target_time(self, stack) -> Tuple[RGBColor, int, RGBColor, int]:
        try:
            color_settings = stack[0]
        except IndexError:
            # no stack
            return self._off_color, -1, self._off_color, -1

        dest_color = color_settings.dest_color
        dest_time = color_settings.dest_time

        # no fade
        if not dest_time:
            # if we are transparent just return the lower layer
            if dest_color is None:
                return self._get_color_and_target_time(stack[1:])
            return dest_color, -1, dest_color, -1

        # fade out
        if dest_color is None:
            _, _, lower_dest_color, lower_dest_time = self._get_color_and_target_time(stack[1:])
            start_time = color_settings.start_time
            if lower_dest_time < 0:
                # no fade going on below current layer
                dest_color = lower_dest_color
            elif start_time < lower_dest_time < dest_time:
                # fade below is shorter than fade out. removing the fade will trigger a new fade in this case
                ratio = (lower_dest_time - dest_time) / (dest_time - start_time)
                dest_color = RGBColor.blend(color_settings.start_color, dest_color, ratio)
                dest_time = lower_dest_time
            else:
                # upper fade is longer. use color target below. this might be slightly inaccurate
                dest_color = lower_dest_color

        # return destination color and time
        return color_settings.start_color, color_settings.start_time, dest_color, dest_time

    # pylint: disable-msg=too-many-return-statements
    def _get_color_and_fade(self, stack, max_fade_ms: int, *, current_time=None) -> Tuple[RGBColor, int, bool]:
        try:
            color_settings = stack[0]
        except IndexError:
            # no stack
            return self._off_color, -1, True

        dest_color = color_settings.dest_color

        # no fade
        if not color_settings.dest_time:
            # if we are transparent just return the lower layer
            if dest_color is None:
                return self._get_color_and_fade(stack[1:], max_fade_ms)
            return dest_color, -1, True

        if current_time is None:
            current_time = self.machine.clock.get_time()

        # fade is done
        if current_time >= color_settings.dest_time:
            # if we are transparent just return the lower layer
            if dest_color is None:
                return self._get_color_and_fade(stack[1:], max_fade_ms)
            return color_settings.dest_color, -1, True

        if dest_color is None:
            dest_color, lower_fade_ms, _ = self._get_color_and_fade(stack[1:], max_fade_ms)
            if lower_fade_ms > 0:
                max_fade_ms = lower_fade_ms

        target_time = current_time + (max_fade_ms / 1000.0)
        # check if fade will be done before max_fade_ms
        if target_time > color_settings.dest_time:
            return dest_color, int((color_settings.dest_time - current_time) * 1000), True

        # figure out the ratio of how far along we are
        try:
            ratio = ((target_time - color_settings.start_time) /
                     (color_settings.dest_time - color_settings.start_time))
        except ZeroDivisionError:
            ratio = 1.0

        return RGBColor.blend(color_settings.start_color, dest_color, ratio), max_fade_ms, False

    def _get_brightness_and_fade(self, max_fade_ms: int, color: str, *, current_time=None) -> Tuple[float, int, bool]:
        uncorrected_color, fade_ms, done = self._get_color_and_fade(self.stack, max_fade_ms, current_time=current_time)
        corrected_color = self.gamma_correct(uncorrected_color)
        corrected_color = self.color_correct(corrected_color)

        if color in ["red", "blue", "green"]:
            brightness = getattr(corrected_color, color) / 255.0
        elif color == "white":
            brightness = min(corrected_color.red, corrected_color.green, corrected_color.blue) / 255.0
        else:
            raise ColorException("Invalid color {}".format(color))
        return brightness, fade_ms, done

    @property
    def _color(self):
        """Getter for color."""
        return self.get_color()

    def get_color_below(self, priority, key):
        """Return an RGBColor() instance of the 'color' setting of the highest color below a certain key.

        Similar to get_color.
        """
        if not self.stack:
            # no stack -> we are black
            return self._off_color

        if self.stack[0].key == key and self.stack[0].priority == priority:
            # fast path for resetting the top element
            return self._get_color_and_fade(self.stack, 0)[0]

        stack = []
        for i, entry in enumerate(self.stack):
            if entry.priority <= priority and entry.key <= key:
                stack = self.stack[i:]
                break
        return self._get_color_and_fade(stack, 0)[0]

    def get_color(self):
        """Return an RGBColor() instance of the 'color' setting of the highest color setting in the stack.

        This is usually the same color as the physical light, but not always (since physical lights are updated once per
        frame, this value could vary.

        Also note the color returned is the "raw" color that does has not had the color correction profile applied.
        """
        return self._get_color_and_fade(self.stack, 0)[0]

    @property
    def fade_in_progress(self) -> bool:
        """Return true if a fade is in progress."""
        return bool(self.stack and self.stack[0].dest_time > self.machine.clock.get_time())



import asyncio

from mpf.tests.MpfTestCase import MpfTestCase, MagicMock, patch


class TestSmbus2(MpfTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/smbus2/'

    def get_platform(self):
        # no force platform. we are testing smbus2
        return False

    def setUp(self):
        smbus = patch('mpf.platforms.smbus2.SMBus2Asyncio')
        self.smbus = smbus.start()
        self.smbus_instance = MagicMock()
        self.smbus.return_value = self.smbus_instance
        self.addCleanup(self.smbus.stop)
        super().setUp()

    def test_i2c(self):
        result = asyncio.Future()
        result.set_result(True)
        self.smbus_instance.write_byte_data = MagicMock(return_value=result)

        device = self.loop.run_until_complete(self.machine.default_platform.configure_i2c("1-17"))

        device.i2c_write8(23, 1337)
        self.machine_run()
        self.smbus_instance.write_byte_data.assert_called_once_with(17, 23, 1337)
        self.smbus.assert_called_once_with('1')

        result = asyncio.Future()
        result.set_result(1337)
        self.smbus_instance.read_byte_data = MagicMock(return_value=result)
        self.assertEqual(1337, self.loop.run_until_complete(device.i2c_read8(23)))
        self.smbus_instance.read_byte_data.assert_called_once_with(17, 23)



"""Contains the base classes for mechanical EM-style score reels."""
import asyncio

from mpf.core.utility_functions import Util

from mpf.core.delays import DelayManager
from mpf.core.system_wide_device import SystemWideDevice


class ScoreReel(SystemWideDevice):

    """Represents an individual electro-mechanical score reel in a pinball machine.

    Multiples reels of this class can be grouped together into ScoreReelGroups
    which collectively make up a display like "Player 1 Score" or "Player 2
    card value", etc.

    This device class is used for all types of mechanical number reels in a
    machine, including reels that have more than ten numbers and that can move
    in multiple directions (such as the credit reel).
    """

    config_section = 'score_reels'
    collection = 'score_reels'
    class_label = 'score_reel'

    def __init__(self, machine, name):
        """Initialise score reel."""
        super().__init__(machine, name)
        self.delay = DelayManager(machine)

        self.value_switches = []
        # This is a list with each element corresponding to a value on the
        # reel. An entry of None means there's no value switch there. An entry
        # of a reference to a switch object means there
        # is a switch there.

        self.assumed_value = -999
        # The assumed value the machine thinks this reel is showing. A value
        # of -999 indicates that the value is unknown.

        self._destination_value = 0
        # Holds the index of the destination the reel is trying to advance to.

        self._runner = None
        # asyncio task which advances the reel

        self._busy = asyncio.Event()
        self._busy.set()
        # will be cleared when the runner is done. set to trigger the runner

        self._ready = asyncio.Event()
        # will be set when this real is ready and shows the destination value

        # stop device on shutdown
        self.machine.events.add_handler("shutdown", self.stop)

    async def _initialize(self):
        await super()._initialize()
        self.log.debug("Configuring score reel with: %s", self.config)

        # figure out how many values we have
        # Add 1 so range is inclusive of the lower limit
        num_values = self.config['limit_hi'] - self.config['limit_lo'] + 1

        self.log.debug("Total reel values: %s", num_values)

        for value in range(num_values):
            self.value_switches.append(self.config.get('switch_' + str(value)))

        self._runner = self.machine.clock.loop.create_task(self._run())
        self._runner.add_done_callback(Util.raise_exceptions)

    def stop(self, **kwargs):
        """Stop device."""
        del kwargs
        if self._runner:
            self._runner.cancel()

    def check_hw_switches(self):
        """Check all the value switches for this score reel.

        This check only happens if `self.ready` is `True`. If the reel is not
        ready, it means another advance request has come in after the initial
        one. In that case then the subsequent advance will call this method
        again when after that advance is done.

        If this method finds an active switch, it sets `self.physical_value` to
        that. Otherwise it sets it to -999. It will also update
        `self.assumed_value` if it finds an active switch. Otherwise it leaves
        that value unchanged.

        This method is automatically called (via a delay) after the reel
        advances. The delay is based on the config value
        `self.config['hw_confirm_time']`.

        TODO: What happens if there are multiple active switches? Currently it
        will return the highest one. Is that ok?
        """
        # check to make sure the 'hw_confirm_time' time has passed. If not then
        # we cannot trust any value we read from the switches

        self.log.debug("Checking hw switches to determine reel value with hw_confirm_time %sms",
                       self.config['hw_confirm_time'])
        for i in range(len(self.value_switches)):
            if self.value_switches[i]:  # not all values have a switch
                if self.machine.switch_controller.is_active(self.value_switches[i],
                                                            ms=self.config['hw_confirm_time']):
                    if self.assumed_value != i:
                        self.log.info("Setting value to %s because that switch is active.", i)
                        if self.assumed_value != -999:
                            self.log.warning("Reel desynced. Assumed: %s. Real: %s", self.assumed_value, i)

                        self.assumed_value = i
                        self._busy.set()
                        self._ready.clear()
                    return

        # check if there is a switch for the current assumed_value
        if (self.assumed_value >= 0 and self.value_switches[self.assumed_value] and
                not self.machine.switch_controller.is_active(self.value_switches[self.assumed_value],
                                                             ms=self.config['hw_confirm_time'])):
            self.log.warning("Resetting value because the switch for %s is not active.", self.assumed_value)
            self.assumed_value = -999
            self._busy.set()
            self._ready.clear()

    async def _run(self):
        await self.machine.events.wait_for_event("init_phase_3")
        self.check_hw_switches()
        while True:
            # wait for either a new value or a switch change
            switch_change_future = self.machine.switch_controller.wait_for_any_switch(
                switches=[switch for switch in self.value_switches if switch is not None], state=2,
                ms=self.config['hw_confirm_time'])
            result = await Util.first([switch_change_future, self._busy.wait()])
            if result == switch_change_future:
                self.check_hw_switches()
                continue

            # advance the reel until we reached our destination position
            await self._advance_reel_if_position_does_not_match()

    async def _advance_reel_if_position_does_not_match(self):
        """Advance reel if the destination value has not been reached."""
        # check if there is any need to do something
        if self._destination_value == self.assumed_value:
            self.log.debug("Reel is already at value %s (will not move)", self._destination_value)
            self._busy.clear()
            self._ready.set()
            return

        self.log.debug("Advancing reel to value %s (current value: %s repeat_pulse_time: %sms)",
                       self._destination_value, self.assumed_value, self.config['repeat_pulse_time'])
        while self._destination_value != self.assumed_value:
            wait_ms = self.config['coil_inc'].pulse(max_wait_ms=500)
            previous_value = self.assumed_value

            await asyncio.sleep((wait_ms + self.config['repeat_pulse_time']) / 1000)
            if self.assumed_value >= 0:
                self.assumed_value += 1
                self.assumed_value %= len(self.value_switches)

            self.check_hw_switches()

            if previous_value != self.assumed_value and self.assumed_value >= 0:
                self.machine.events.post('reel_{}_advance'.format(self.name))
                '''event: reel_(name)_advance

                desc: The reel (name) advanced to the next position.
                '''
            self.log.debug("Assumed value: %s", self.assumed_value)

        self._busy.clear()
        self._ready.set()
        self.log.debug("Advancing to %s successful.", self._destination_value)

    def wait_for_ready(self):
        """Return a future for ready."""
        return self._ready.wait()

    def set_destination_value(self, value):
        """Return the integer value of the destination this reel is moving to.

        Args:
        ----
            value: Destination value which this reel should try to reach.

        Returns: The value of the destination. If the current
            `self.assumed_value` is -999, this method will always return -999
            since it doesn't know where the reel is and therefore doesn't know
            what the destination value would be.
        """
        if self._destination_value != value:
            self.log.debug("Setting new score_reel value. Old destination value: %s, New destination value: %s",
                           self._destination_value, value)

            self._destination_value = value

            self._busy.set()
            self._ready.clear()



from mpf.tests.MpfFakeGameTestCase import MpfFakeGameTestCase
from mpf.tests.MpfTestCase import test_config


class TestShotGroups(MpfFakeGameTestCase):

    def get_config_file(self):
        return 'test_shot_groups.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/shots/'

    def test_disabled_when_no_game(self):
        # all shot group functionality should be disabled if there is not a
        # game in progress. Really we're just making sure this doesn't crash.

        self.machine.events.post('s_rotate_l_active')
        self.advance_time_and_run()

        self.hit_and_release_switch("switch_1")
        self.hit_and_release_switch("switch_2")
        self.hit_and_release_switch("switch_3")
        self.hit_and_release_switch("switch_4")

    def test_events_and_complete(self):
        self.mock_event("test_group_complete")
        self.mock_event("test_group_hit")

        self.start_game()

        # it should post events. here for the initial profile state (on mode start)
        self.assertEventNotCalled("test_group_hit")
        self.assertEventCalledWith("test_group_complete", state="unlit")
        self.mock_event("test_group_complete")
        self.mock_event("test_group_hit")
        self.assertPlaceholderEvaluates("unlit", "device.shot_groups.test_group.common_state")

        self.hit_and_release_switch("switch_1")

        # it posts nothing because the the there is no common state
        self.assertEventCalled("test_group_hit")
        self.assertEventNotCalled("test_group_complete")

        self.hit_and_release_switch("switch_2")
        self.assertPlaceholderEvaluates(None, "device.shot_groups.test_group.common_state")
        self.hit_and_release_switch("switch_3")
        self.hit_and_release_switch("switch_4")

        self.assertEventCalled("test_group_hit")
        self.assertEventCalledWith("test_group_complete", state="lit")
        self.assertPlaceholderEvaluates("lit", "device.shot_groups.test_group.common_state")

        self.stop_game()

    @test_config("test_shot_group_rotate_with_exclude.yaml")
    def test_rotate_state_names_to_not_rotate(self):
        """Test state_names_to_not_rotate in profile."""
        self.start_game()

        # unlit cannot rotate. all other states can
        self.assertEqual("unlit", self.machine.shots["shot_1"].state_name)
        self.assertEqual("unlit", self.machine.shots["shot_2"].state_name)
        self.assertEqual("unlit", self.machine.shots["shot_3"].state_name)
        self.assertEqual("unlit", self.machine.shots["shot_4"].state_name)

        self.hit_and_release_switch("s_rotate_r")

        # nothing should happen but it should also not crash
        self.assertEqual("unlit", self.machine.shots["shot_1"].state_name)
        self.assertEqual("unlit", self.machine.shots["shot_2"].state_name)
        self.assertEqual("unlit", self.machine.shots["shot_3"].state_name)
        self.assertEqual("unlit", self.machine.shots["shot_4"].state_name)

        self.hit_and_release_switch("switch_1")

        # first shot is advanced
        self.assertEqual("red", self.machine.shots["shot_1"].state_name)
        self.assertEqual("unlit", self.machine.shots["shot_2"].state_name)
        self.assertEqual("unlit", self.machine.shots["shot_3"].state_name)
        self.assertEqual("unlit", self.machine.shots["shot_4"].state_name)

        self.hit_and_release_switch("s_rotate_r")

        # nothing should happen but it should also not crash
        self.assertEqual("red", self.machine.shots["shot_1"].state_name)
        self.assertEqual("unlit", self.machine.shots["shot_2"].state_name)
        self.assertEqual("unlit", self.machine.shots["shot_3"].state_name)
        self.assertEqual("unlit", self.machine.shots["shot_4"].state_name)

        self.hit_and_release_switch("switch_2")
        self.hit_and_release_switch("switch_3")
        self.hit_and_release_switch("switch_3")

        # some more shots to advance
        self.assertEqual("red", self.machine.shots["shot_1"].state_name)
        self.assertEqual("red", self.machine.shots["shot_2"].state_name)
        self.assertEqual("orange", self.machine.shots["shot_3"].state_name)
        self.assertEqual("unlit", self.machine.shots["shot_4"].state_name)

        self.hit_and_release_switch("s_rotate_r")

        # skip the unlit shot
        self.assertEqual("orange", self.machine.shots["shot_1"].state_name)
        self.assertEqual("red", self.machine.shots["shot_2"].state_name)
        self.assertEqual("red", self.machine.shots["shot_3"].state_name)
        self.assertEqual("unlit", self.machine.shots["shot_4"].state_name)


    def test_rotate(self):
        self.start_game()

        self.assertEqual("unlit", self.machine.shots["shot_1"].state_name)
        self.assertEqual("unlit", self.machine.shots["shot_2"].state_name)
        self.assertEqual("unlit", self.machine.shots["shot_3"].state_name)
        self.assertEqual("unlit", self.machine.shots["shot_4"].state_name)

        self.hit_and_release_switch("switch_1")

        self.assertEqual("lit", self.machine.shots["shot_1"].state_name)
        self.assertEqual("unlit", self.machine.shots["shot_2"].state_name)
        self.assertEqual("unlit", self.machine.shots["shot_3"].state_name)
        self.assertEqual("unlit", self.machine.shots["shot_4"].state_name)

        self.hit_and_release_switch("s_rotate_r")

        self.assertEqual("unlit", self.machine.shots["shot_1"].state_name)
        self.assertEqual("lit", self.machine.shots["shot_2"].state_name)
        self.assertEqual("unlit", self.machine.shots["shot_3"].state_name)
        self.assertEqual("unlit", self.machine.shots["shot_4"].state_name)

        self.hit_and_release_switch("switch_1")

        self.assertEqual("lit", self.machine.shots["shot_1"].state_name)
        self.assertEqual("lit", self.machine.shots["shot_2"].state_name)
        self.assertEqual("unlit", self.machine.shots["shot_3"].state_name)
        self.assertEqual("unlit", self.machine.shots["shot_4"].state_name)

        self.hit_and_release_switch("s_rotate_r")

        self.assertEqual("unlit", self.machine.shots["shot_1"].state_name)
        self.assertEqual("lit", self.machine.shots["shot_2"].state_name)
        self.assertEqual("lit", self.machine.shots["shot_3"].state_name)
        self.assertEqual("unlit", self.machine.shots["shot_4"].state_name)

        self.hit_and_release_switch("s_rotate_r")

        self.assertEqual("unlit", self.machine.shots["shot_1"].state_name)
        self.assertEqual("unlit", self.machine.shots["shot_2"].state_name)
        self.assertEqual("lit", self.machine.shots["shot_3"].state_name)
        self.assertEqual("lit", self.machine.shots["shot_4"].state_name)

        self.hit_and_release_switch("s_rotate_r")

        self.assertEqual("lit", self.machine.shots["shot_1"].state_name)
        self.assertEqual("unlit", self.machine.shots["shot_2"].state_name)
        self.assertEqual("unlit", self.machine.shots["shot_3"].state_name)
        self.assertEqual("lit", self.machine.shots["shot_4"].state_name)

        self.hit_and_release_switch("s_rotate_l")

        self.assertEqual("unlit", self.machine.shots["shot_1"].state_name)
        self.assertEqual("unlit", self.machine.shots["shot_2"].state_name)
        self.assertEqual("lit", self.machine.shots["shot_3"].state_name)
        self.assertEqual("lit", self.machine.shots["shot_4"].state_name)

    def test_profile_from_shot(self):
        self.start_game()
        self.advance_time_and_run()

        self.assertLightColor("led_30", 'off')
        self.assertLightColor("led_31", 'off')

        self.hit_and_release_switch('switch_30')
        self.advance_time_and_run()
        self.assertLightColor("led_30", 'red')
        self.assertLightColor("led_31", 'off')

        self.hit_and_release_switch('switch_30')
        self.advance_time_and_run()
        self.assertLightColor("led_30", 'orange')
        self.assertLightColor("led_31", 'off')

        self.hit_and_release_switch('switch_31')
        self.advance_time_and_run()
        self.assertLightColor("led_30", 'orange')
        self.assertLightColor("led_31", 'red')

    def test_control_events(self):
        # tests control events at the shot_group level

        shot32 = self.machine.shots["shot_32"]
        shot33 = self.machine.shots["shot_33"]
        group32 = self.machine.shot_groups["shot_group_32"]

        self.mock_event("shot_32_hit")
        self.mock_event("shot_33_hit")
        self.mock_event("shot_group_32_hit")

        self.start_game()

        # Since this shot has custom enable events, it should not be enabled on
        # game start
        self.assertFalse(shot32.enabled)
        self.assertFalse(shot33.enabled)

        # test enabling via event
        self.machine.events.post('group32_enable')
        self.advance_time_and_run()

        self.assertTrue(shot32.enabled)
        self.assertTrue(shot33.enabled)

        # test disabling via event
        self.machine.events.post('group32_disable')
        self.advance_time_and_run()

        self.assertFalse(shot32.enabled)
        self.assertFalse(shot33.enabled)

        # test enabling via event
        self.machine.events.post('group32_enable')
        self.advance_time_and_run()

        self.assertTrue(shot32.enabled)
        self.assertTrue(shot33.enabled)

        # test rotate without rotation enabled
        shot32.advance()
        self.advance_time_and_run()
        self.assertEqual(shot32.state_name, 'red')
        self.assertEqual(shot33.state_name, 'unlit')
        self.assertLightColor("led_32", 'red')
        self.assertLightColor("led_33", 'off')
        self.assertFalse(group32.rotation_enabled)
        self.assertPlaceholderEvaluates(False, "device.shot_groups.shot_group_32.rotation_enabled")

        self.machine.events.post('group32_rotate')
        self.advance_time_and_run()
        self.assertEqual(shot32.state_name, 'red')
        self.assertEqual(shot33.state_name, 'unlit')
        self.assertLightColor("led_32", 'red')
        self.assertLightColor("led_33", 'off')

        # test rotation enable
        self.machine.events.post('group32_enable_rotation')
        self.advance_time_and_run()
        self.assertTrue(group32.rotation_enabled)
        self.assertPlaceholderEvaluates(True, "device.shot_groups.shot_group_32.rotation_enabled")

        # test that rotate works now
        self.machine.events.post('group32_rotate_left')
        self.advance_time_and_run()
        self.assertEqual(shot32.state_name, 'unlit')
        self.assertEqual(shot33.state_name, 'red')
        self.assertLightColor("led_32", 'off')
        self.assertLightColor("led_33", 'red')

        # test disable rotation
        self.machine.events.post('group32_disable_rotation')
        self.advance_time_and_run()
        self.assertFalse(group32.rotation_enabled)

        # test that rotate works now
        self.machine.events.post('group32_rotate_left')
        self.advance_time_and_run()

        # test that rotate did not happen
        self.assertEqual(shot32.state_name, 'unlit')
        self.assertEqual(shot33.state_name, 'red')
        self.assertLightColor("led_32", 'off')
        self.assertLightColor("led_33", 'red')

        # test reset
        # test enabling via event
        self.machine.events.post('group32_reset')
        self.advance_time_and_run()

        self.assertTrue(shot32.enabled)
        self.assertTrue(shot33.enabled)
        self.assertEqual(shot32.state_name, 'unlit')
        self.assertEqual(shot33.state_name, 'unlit')
        self.assertLightColor("led_32", 'off')
        self.assertLightColor("led_33", 'off')

        # test restart
        # first advance and disable
        shot32.advance()
        shot33.advance()
        shot32.disable()
        shot33.disable()
        self.assertFalse(shot32.enabled)
        self.assertFalse(shot33.enabled)
        self.assertEqual(shot32.state_name, 'red')
        self.assertEqual(shot33.state_name, 'red')
        # ensure all shots are enabled and at the first state
        self.machine.events.post('group32_restart')
        self.advance_time_and_run()
        self.assertTrue(shot32.enabled)
        self.assertTrue(shot33.enabled)
        self.assertEqual(shot32.state_name, 'unlit')
        self.assertEqual(shot33.state_name, 'unlit')

    def test_rotation_pattern(self):
        shot40 = self.machine.shots["shot_40"]
        shot41 = self.machine.shots["shot_41"]
        shot42 = self.machine.shots["shot_42"]
        group40 = self.machine.shot_groups["shot_group_40"]

        self.start_game()

        shot40.advance()
        self.advance_time_and_run()
        self.assertEqual(shot40.state_name, 'red')
        self.assertEqual(shot41.state_name, 'unlit')
        self.assertEqual(shot42.state_name, 'unlit')
        self.assertLightColor("led_40", 'red')
        self.assertLightColor("led_41", 'off')
        self.assertLightColor("led_42", 'off')

        group40.rotate()
        self.advance_time_and_run()
        self.assertEqual(shot40.state_name, 'unlit')
        self.assertEqual(shot41.state_name, 'red')
        self.assertEqual(shot42.state_name, 'unlit')
        self.assertLightColor("led_40", 'off')
        self.assertLightColor("led_41", 'red')
        self.assertLightColor("led_42", 'off')

        group40.rotate()
        self.advance_time_and_run()
        self.assertEqual(shot40.state_name, 'unlit')
        self.assertEqual(shot41.state_name, 'unlit')
        self.assertEqual(shot42.state_name, 'red')
        self.assertLightColor("led_40", 'off')
        self.assertLightColor("led_41", 'off')
        self.assertLightColor("led_42", 'red')

        group40.rotate()
        self.advance_time_and_run()
        self.assertEqual(shot40.state_name, 'unlit')
        self.assertEqual(shot41.state_name, 'red')
        self.assertEqual(shot42.state_name, 'unlit')
        self.assertLightColor("led_40", 'off')
        self.assertLightColor("led_41", 'red')
        self.assertLightColor("led_42", 'off')

        group40.rotate()
        self.advance_time_and_run()
        self.assertEqual(shot40.state_name, 'red')
        self.assertEqual(shot41.state_name, 'unlit')
        self.assertEqual(shot42.state_name, 'unlit')
        self.assertLightColor("led_40", 'red')
        self.assertLightColor("led_41", 'off')
        self.assertLightColor("led_42", 'off')

        group40.rotate()
        self.advance_time_and_run()
        self.assertEqual(shot40.state_name, 'unlit')
        self.assertEqual(shot41.state_name, 'red')
        self.assertEqual(shot42.state_name, 'unlit')
        self.assertLightColor("led_40", 'off')
        self.assertLightColor("led_41", 'red')
        self.assertLightColor("led_42", 'off')

    def test_profile_on_second_ball(self):
        self.start_game()

        self.assertLightChannel("l_special_left", 0)
        self.assertLightChannel("l_special_right", 0)

        shot = self.machine.shots["lane_special_left"]

        self.assertEqual('unlit_toggle', shot.state_name)

        # toggle on
        self.hit_and_release_switch("s_special_left")
        self.advance_time_and_run(.1)
        self.assertLightChannel("l_special_left", 255)
        self.assertEqual('lit_toggle', shot.state_name)

        # toggle off
        self.hit_and_release_switch("s_special_left")
        self.advance_time_and_run(.1)
        self.assertLightChannel("l_special_left", 0)
        self.assertEqual('unlit_toggle', shot.state_name)

        # drain ball and try on the second ball
        self.drain_all_balls()
        self.assertBallNumber(2)

        self.assertEqual('unlit_toggle', shot.state_name)

        # toggle on
        self.hit_and_release_switch("s_special_left")
        self.advance_time_and_run(.1)
        self.assertLightChannel("l_special_left", 255)
        self.assertEqual('lit_toggle', shot.state_name)

        # toggle off
        self.hit_and_release_switch("s_special_left")
        self.advance_time_and_run(.1)
        self.assertLightChannel("l_special_left", 0)
        self.assertEqual('unlit_toggle', shot.state_name)

        self.drain_all_balls()
        self.assertBallNumber(3)

        # toggle on
        self.hit_and_release_switch("s_special_left")
        self.advance_time_and_run(.1)
        self.assertLightChannel("l_special_left", 255)
        self.assertEqual('lit_toggle', shot.state_name)

        # toggle off
        self.hit_and_release_switch("s_special_left")
        self.advance_time_and_run(.1)
        self.assertLightChannel("l_special_left", 0)
        self.assertEqual('unlit_toggle', shot.state_name)

        # toggle on
        self.hit_and_release_switch("s_special_left")
        self.advance_time_and_run(.1)
        self.assertLightChannel("l_special_left", 255)
        self.assertEqual('lit_toggle', shot.state_name)

        self.drain_all_balls()
        self.assertGameIsNotRunning()

        # shot should turn off after game
        self.assertLightChannel("l_special_left", 0)



"""Contains the base class for diverter devices."""
from collections import deque

from mpf.core.events import event_handler

from mpf.core.delays import DelayManager
from mpf.core.device_monitor import DeviceMonitor
from mpf.core.system_wide_device import SystemWideDevice


@DeviceMonitor("active", "enabled", "eject_state")
class Diverter(SystemWideDevice):

    """Represents a diverter in a pinball machine.

    Args: Same as the Device parent class.
    """

    __slots__ = ["delay", "active", "enabled", "diverting_ejects_count", "eject_state", "eject_attempt_queue"]

    config_section = 'diverters'
    collection = 'diverters'
    class_label = 'diverter'

    def __init__(self, machine, name):
        """Initialise diverter."""
        super().__init__(machine, name)

        self.delay = DelayManager(machine)

        # Attributes
        self.active = False
        self.enabled = False

        self.diverting_ejects_count = 0
        self.eject_state = False
        self.eject_attempt_queue = deque()

    async def _initialize(self):
        await super()._initialize()
        # register for feeder device eject events
        for feeder_device in self.config['feeder_devices']:
            self.machine.events.add_handler(
                'balldevice_' + feeder_device.name +
                '_ball_eject_attempt',
                self._feeder_eject_attempt)

            self.machine.events.add_handler(
                'balldevice_' + feeder_device.name +
                '_ejecting_ball',
                self._feeder_ejecting)

            self.machine.events.add_handler(
                'balldevice_' + feeder_device.name +
                '_ball_eject_failed',
                self._feeder_eject_count_decrease)

            self.machine.events.add_handler(
                'balldevice_' + feeder_device.name +
                '_ball_eject_success',
                self._feeder_eject_count_decrease)

        self.machine.events.add_handler('init_phase_3',
                                        self._register_switches)

        if self.config['ball_search_order']:
            self.config['playfield'].ball_search.register(
                self.config['ball_search_order'], self._ball_search, self.name,
                restore_callback=self._ball_search_restore)

    def _register_switches(self, **kwargs):
        del kwargs
        # register for deactivation switches
        for switch in self.config['deactivation_switches']:
            self.machine.switch_controller.add_switch_handler_obj(
                switch, self.deactivate)

        # register for disable switches:
        for switch in self.config['disable_switches']:
            self.machine.switch_controller.add_switch_handler_obj(
                switch, self.disable)

    @event_handler(1)
    def event_reset(self, **kwargs):
        """Handle reset control event."""
        del kwargs
        self.reset()

    def reset(self):
        """Reset and deactivate the diverter."""
        self.deactivate()

    @event_handler(10)
    def event_enable(self, auto=False, **kwargs):
        """Handle enable control event."""
        del kwargs
        self.enable(auto)

    def enable(self, auto=False):
        """Enable this diverter.

        Args:
        ----
            auto: Boolean value which is used to indicate whether this
                  diverter enabled itself automatically. This is passed to the
                  event which is posted.

        If an 'activation_switches' is configured, then this method writes a
        hardware autofire rule to the pinball controller which fires the
        diverter coil when the switch is activated.

        If no `activation_switches` is specified, then the diverter is activated
        immediately.
        """
        if self.enabled:
            return
        self.enabled = True

        self.machine.events.post('diverter_' + self.name + '_enabling',
                                 auto=auto)
        '''event: diverter_(name)_enabling
        desc: The diverter called (name) is enabling itself. Note that if this
            diverter has ``activation_switches:`` configured, it will not
            physically activate until one of those switches is hit. Otherwise
            this diverter will activate immediately.

        args:
            auto: Boolean which indicates whether this diverter enabled itself
                automatically for the purpose of routing balls to their proper
                location(s).
        '''

        if self.config['activation_switches']:
            self._enable_switches()
        elif self.config['activate_events']:
            pass
        else:
            self.activate()

    @event_handler(0)
    def event_disable(self, auto=False, **kwargs):
        """Handle disable control event."""
        del kwargs
        self.disable(auto)

    def disable(self, auto=False):
        """Disable this diverter.

        This method will remove the hardware rule if this diverter is activated
        via a hardware switch.

        Args:
        ----
            auto: Boolean value which is used to indicate whether this
                diverter disabled itself automatically. This is passed to the
                event which is posted.
            **kwargs: This is here because this disable method is called by
                whatever event the game programmer specifies in their machine
                configuration file, so we don't know what event that might be
                or whether it has random kwargs attached to it.
        """
        if not self.enabled:
            return
        self.enabled = False

        self.machine.events.post('diverter_' + self.name + '_disabling',
                                 auto=auto)
        '''event: diverter_(name)_disabling
        desc: The diverter called (name) is disabling itself. Note that if this
            diverter has ``activation_switches:`` configured, it will not
            physically deactivate now, instead deactivating based on switch
            hits and timing. Otherwise this diverter will deactivate immediately.

        args:
            auto: Boolean which indicates whether this diverter disabled itself
                automatically for the purpose of routing balls to their proper
                location(s).
        '''

        self.debug_log("Disabling Diverter")
        if self.config['activation_switches']:
            self._disable_switches()
        # if there is no deactivation way
        if not (self.config['activation_time'] or self.config['deactivation_switches'] or
           self.config['deactivate_events']):
            self.deactivate()

    def _coil_activate(self):
        """Activate the coil."""
        if self.config['activation_coil']:
            if self.config['type'] == 'pulse':
                self.config['activation_coil'].pulse()
            elif self.config['type'] == 'hold':
                self.config['activation_coil'].enable()

    def _coil_deactivate(self):
        """Deactivate the coil."""
        if self.config['activation_coil']:
            self.config['activation_coil'].disable()

        if self.config['deactivation_coil']:
            self.config['deactivation_coil'].pulse()

    @event_handler(9)
    def event_activate(self, **kwargs):
        """Handle activate control event."""
        del kwargs
        if not self.enabled:
            return
        self.activate()

    def activate(self):
        """Physically activate this diverter's coil."""
        self.debug_log("Activating Diverter")
        self.active = True

        self.machine.events.post('diverter_' + self.name + '_activating')
        '''event: diverter_(name)_activating
        desc: The diverter called (name) is activating itself, which means
            it's physically pulsing or holding the coil to move.

        '''
        self._coil_activate()
        self.schedule_deactivation()

    @event_handler(2)
    def event_deactivate(self, **kwargs):
        """Handle deactivate control event."""
        del kwargs
        if not self.enabled:
            return
        self.deactivate()

    def deactivate(self):
        """Deactivate this diverter.

        This method will disable the activation_coil, and (optionally) if it's
        configured with a deactivation coil, it will pulse it.
        """
        self.debug_log("Deactivating Diverter")
        self.active = False

        if self.config['activation_time']:
            self.delay.remove('deactivate_timed')

        self.machine.events.post('diverter_' + self.name + '_deactivating')
        '''event: diverter_(name)_deactivating
        desc: The diverter called (name) is deativating itself.

        '''
        self._coil_deactivate()

    def schedule_deactivation(self):
        """Schedule a delay to deactivate this diverter."""
        if self.config['activation_time']:
            self.delay.add(name='deactivate_timed', ms=self.config['activation_time'],
                           callback=self.deactivate)

    def _enable_switches(self):
        """Register switch handler on activation switches."""
        self.debug_log("Enabling Diverter sw switches: %s",
                       self.config['activation_switches'])

        for switch in self.config['activation_switches']:
            self.machine.switch_controller.add_switch_handler_obj(
                switch=switch, callback=self.activate)

    def _disable_switches(self):
        """Deregister switch handlers for activation switches."""
        self.debug_log("Disabling Diverter sw switches: %s",
                       self.config['activation_switches'])

        for switch in self.config['activation_switches']:
            self.machine.switch_controller.remove_switch_handler(
                switch_name=switch.name, callback=self.activate)

    def _feeder_eject_count_decrease(self, target, **kwargs):
        del target
        del kwargs
        self.debug_log("Source reported success")
        if self.config['cool_down_time']:
            self.delay.add(self.config['cool_down_time'], self._reduce_eject_count)
        else:
            self._reduce_eject_count()

    def _reduce_eject_count(self):
        self.diverting_ejects_count -= 1
        if self.diverting_ejects_count <= 0:
            self.diverting_ejects_count = 0

            # If there are ejects waiting for the other target switch diverter
            if self.eject_attempt_queue:
                # And perform those ejects
                if self.config['allow_multiple_concurrent_ejects_to_same_side']:
                    while self.eject_attempt_queue:
                        self.diverting_ejects_count += 1
                        queue = self.eject_attempt_queue.pop()
                        queue.clear()
                else:
                    if self.eject_attempt_queue:
                        self.diverting_ejects_count += 1
                        queue = self.eject_attempt_queue.pop()
                        queue.clear()
            elif self.active and not self.config['activation_time']:
                # if diverter is active and no more ejects are ongoing
                self.deactivate()

    def _get_desired_state(self, target):
        desired_state = None
        if target in self.config['targets_when_active']:
            desired_state = True

        elif target in self.config['targets_when_inactive']:
            desired_state = False
        return desired_state

    def _feeder_eject_attempt(self, queue, target, **kwargs):
        # Event handler which is called when one of this diverter's feeder
        # devices attempts to eject a ball. This is what allows this diverter
        # to get itself in the right position to send the ball to where it needs
        # to go.

        # Since the 'target' kwarg is going to be an object, not a name, we need
        # to figure out if this object is one of the targets of this diverter.
        del kwargs
        self.debug_log("Feeder device eject attempt for target: %s", target)

        desired_state = self._get_desired_state(target)

        if desired_state is None:
            self.debug_log("Feeder device ejects to an unknown target: %s. "
                           "Ignoring!", target.name)
            return

        if self.diverting_ejects_count > 0:
            if self.config['allow_multiple_concurrent_ejects_to_same_side'] and self.eject_state != desired_state:
                self.debug_log("Feeder devices tries to eject to a target which "
                               "would require a state change. Postponing that "
                               "because we have an eject to the other side")
                queue.wait()
                self.eject_attempt_queue.append(queue)
                return
            if not self.config['allow_multiple_concurrent_ejects_to_same_side']:
                self.debug_log("More than one eject and allow_multiple_concurrent_ejects_to_same_side is false")
                queue.wait()
                self.eject_attempt_queue.append(queue)
                return

        self.diverting_ejects_count += 1
        self.eject_state = desired_state

    def _feeder_ejecting(self, target, **kwargs):
        """Enable or disable diverter on eject."""
        del kwargs
        self.debug_log("Feeder device is ejecting for target: %s", target)

        desired_state = self._get_desired_state(target)

        if desired_state is None:
            self.debug_log("Feeder device ejects to an unknown target: %s. "
                           "Ignoring!", target.name)
            return

        if desired_state:
            self.debug_log("Enabling diverter since eject target is on the "
                           "active target list")
            self.enable()
        elif not desired_state:
            self.debug_log("Disabling diverter since eject target is on the "
                           "inactive target list")
            self.disable()

    def _ball_search(self, phase, iteration):
        del phase
        del iteration
        self._coil_activate()
        self.machine.delay.add(self.config['ball_search_hold_time'],
                               self._coil_deactivate,
                               'diverter_{}_ball_search'.format(self.name))
        return True

    def _ball_search_restore(self):
        """Restore state after ball search ended."""
        if self.active:
            self._coil_activate()
        else:
            self._coil_deactivate()



from mpf.tests.MpfTestCase import MpfTestCase
from unittest.mock import MagicMock


class TestBallDeviceTriggerEvents(MpfTestCase):
    def __init__(self, test_map):
        super().__init__(test_map)
        self._captured = 0
        self._enter = 0
        self._missing = 0
        self._requesting = 0
        self._queue = False

    def get_config_file(self):
        return 'test_ball_device_trigger_events.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/ball_device/'

    def _missing_ball(self, **kwargs):
        del kwargs
        self._missing += 1

    def _requesting_ball(self, balls, **kwargs):
        del kwargs
        self._requesting += balls

    def _ball_enter(self, new_balls, unclaimed_balls, **kwargs):
        del unclaimed_balls
        del kwargs
        self._enter += new_balls

    def _captured_from_pf(self, balls, **kwargs):
        del kwargs
        self._captured += balls

    def test_manual_successful_eject_to_pf(self):
        coil1 = self.machine.coils['eject_coil1']
        coil2 = self.machine.coils['eject_coil2']
        device1 = self.machine.ball_devices['test_trough']
        device2 = self.machine.ball_devices['test_launcher']
        playfield = self.machine.ball_devices['playfield']

        self.machine.events.add_handler('balldevice_captured_from_playfield',
                                        self._captured_from_pf)
        self.machine.events.add_handler('balldevice_ball_missing',
                                        self._missing_ball)
        self._enter = 0
        self._captured = 0
        self._missing = 0

        # add an initial ball to trough
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.advance_time_and_run(1)
        self.assertEqual(1, self._captured)
        self._captured = 0
        self.assertEqual(0, playfield.balls)

        # it should keep the ball
        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()
        self.assertEqual(1, device1.balls)
        assert not coil1.pulse.called
        assert not coil2.pulse.called

        # request an ball
        playfield.add_ball(player_controlled=True)
        self.advance_time_and_run(1)

        # trough eject
        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called

        self.machine.switch_controller.process_switch("s_ball_switch1", 0)
        self.advance_time_and_run(1)
        self.assertEqual(0, device1.balls)

        # launcher receives but waits for player to eject
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      1)
        self.advance_time_and_run(1)
        self.assertEqual(1, device2.balls)

        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called

        # player shoots the ball
        self.machine.switch_controller.process_switch("s_launch", 1)
        self.advance_time_and_run(1)
        self.machine.switch_controller.process_switch("s_launch", 0)
        self.advance_time_and_run(1)

        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      0)
        self.advance_time_and_run(1)
        self.assertEqual(0, device2.balls)

        self.machine.switch_controller.process_switch("s_playfield", 1)
        self.advance_time_and_run(0.1)
        self.machine.switch_controller.process_switch("s_playfield", 0)
        self.advance_time_and_run(1)

        self.assertEqual(1, playfield.balls)
        self.assertEqual(0, self._captured)
        self.assertEqual(0, self._missing)

    def test_eject_without_trigger_events(self):
        coil1 = self.machine.coils['eject_coil1']
        coil2 = self.machine.coils['eject_coil2']
        device1 = self.machine.ball_devices['test_trough']
        device2 = self.machine.ball_devices['test_launcher']
        playfield = self.machine.ball_devices['playfield']

        self.machine.events.add_handler('balldevice_captured_from_playfield',
                                        self._captured_from_pf)
        self.machine.events.add_handler('balldevice_ball_missing',
                                        self._missing_ball)
        self._enter = 0
        self._captured = 0
        self._missing = 0

        # add an initial ball to trough
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.advance_time_and_run(1)
        self.assertEqual(1, self._captured)
        self._captured = 0
        self.assertEqual(0, playfield.balls)

        # it should keep the ball
        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()
        self.assertEqual(1, device1.balls)
        assert not coil1.pulse.called
        assert not coil2.pulse.called

        # request an ball without player_controlled eject
        playfield.add_ball(player_controlled=False)
        self.advance_time_and_run(1)

        # trough eject
        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called

        self.machine.switch_controller.process_switch("s_ball_switch1", 0)
        self.advance_time_and_run(1)
        self.assertEqual(0, device1.balls)

        # launcher receives and should just eject
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      1)
        self.advance_time_and_run(1)
        self.assertEqual(1, device2.balls)

        self.assertTrue(coil1.pulse.called)
        self.assertTrue(coil2.pulse.called)
        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()

        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      0)
        self.advance_time_and_run(1)
        self.assertEqual(0, device2.balls)

        self.machine.switch_controller.process_switch("s_playfield", 1)
        self.advance_time_and_run(0.1)
        self.machine.switch_controller.process_switch("s_playfield", 0)
        self.advance_time_and_run(1)

        self.assertEqual(1, playfield.balls)

        self.assertEqual(0, self._captured)
        self.assertEqual(0, self._missing)

    def test_manual_with_retry_to_pf(self):
        coil1 = self.machine.coils['eject_coil1']
        coil2 = self.machine.coils['eject_coil2']
        device1 = self.machine.ball_devices['test_trough']
        device2 = self.machine.ball_devices['test_launcher']
        playfield = self.machine.ball_devices['playfield']

        self.machine.events.add_handler('balldevice_captured_from_playfield',
                                        self._captured_from_pf)
        self.machine.events.add_handler('balldevice_ball_missing',
                                        self._missing_ball)
        self._enter = 0
        self._captured = 0
        self._missing = 0

        # add an initial ball to trough
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.advance_time_and_run(1)
        self.assertEqual(1, self._captured)
        self._captured = 0
        self.assertEqual(0, playfield.balls)

        # it should keep the ball
        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()
        self.assertEqual(1, device1.balls)
        assert not coil1.pulse.called
        assert not coil2.pulse.called

        # request an ball
        playfield.add_ball(player_controlled=True)
        self.advance_time_and_run(1)

        # trough eject
        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called

        self.machine.switch_controller.process_switch("s_ball_switch1", 0)
        self.advance_time_and_run(1)
        self.assertEqual(0, device1.balls)

        # launcher receives but waits for player to eject
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      1)
        self.advance_time_and_run(1)
        self.assertEqual(1, device2.balls)

        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called

        # player shoots the ball
        self.machine.switch_controller.process_switch("s_launch", 1)
        self.advance_time_and_run(1)
        self.machine.switch_controller.process_switch("s_launch", 0)
        self.advance_time_and_run(1)

        self.assertTrue(coil1.pulse.called)
        self.assertTrue(coil2.pulse.called)
        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()

        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      0)
        self.advance_time_and_run(1)
        self.assertEqual(0, device2.balls)

        self.advance_time_and_run(3)

        # too soft and it comes back
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      1)
        self.advance_time_and_run(3)
        self.assertEqual(1, device2.balls)

        # player drinks his coffee
        self.advance_time_and_run(300)

        assert not coil1.pulse.called
        assert not coil2.pulse.called

        # player shoots the ball again
        self.machine.switch_controller.process_switch("s_launch", 1)
        self.advance_time_and_run(1)
        self.machine.switch_controller.process_switch("s_launch", 0)
        self.advance_time_and_run(1)

        assert not coil1.pulse.called
        self.assertTrue(coil2.pulse.called)

        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      0)
        self.advance_time_and_run(1)
        self.assertEqual(0, device2.balls)

        assert not coil1.pulse.called
        self.assertTrue(coil2.pulse.called)

        self.machine.switch_controller.process_switch("s_playfield", 1)
        self.advance_time_and_run(0.1)
        self.machine.switch_controller.process_switch("s_playfield", 0)
        self.advance_time_and_run(1)

        self.assertEqual(1, playfield.balls)

        self.assertEqual(0, self._captured)
        self.assertEqual(0, self._missing)



from mpf.platforms.interfaces.driver_platform_interface import PulseSettings, HoldSettings

from mpf.core.platform import SwitchSettings, DriverSettings

from mpf.tests.MpfTestCase import MpfTestCase
from unittest.mock import MagicMock, call


class TestFlippersHoldNoEos(MpfTestCase):

    def get_config_file(self):
        return 'hold_no_eos.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/flippers/'

    def get_platform(self):
        return 'virtual'

    def test_hold_no_eos(self):
        self.machine.default_platform.set_pulse_on_hit_and_release_rule = MagicMock()
        self.machine.default_platform.set_pulse_on_hit_and_enable_and_release_rule = MagicMock()

        # Config uses enable_events to enable the flipper on boot
        self.assertTrue(self.machine.flippers["left_flipper"]._enabled)
        self.machine.flippers["left_flipper"].disable()

        self.assertFalse(self.machine.flippers["left_flipper"]._enabled)
        self.machine.flippers["left_flipper"].enable()
        #     def set_hw_rule(self, sw_name, sw_activity, driver_name, driver_action,
        #                     disable_on_release=True, drive_now=False,
        #                     **driver_settings_overrides):
        self.assertEqual(1, len(self.machine.default_platform.set_pulse_on_hit_and_release_rule._mock_call_args_list))
        self.assertEqual(1, len(self.machine.default_platform.set_pulse_on_hit_and_enable_and_release_rule.
                                _mock_call_args_list))

        self.machine.default_platform.set_pulse_on_hit_and_release_rule.assert_called_once_with(
            SwitchSettings(hw_switch=self.machine.switches["s_left_flipper"].hw_switch, invert=False, debounce=False),
            DriverSettings(hw_driver=self.machine.coils["c_flipper_left_main"].hw_driver,
                           pulse_settings=PulseSettings(power=1.0, duration=30),
                           hold_settings=None, recycle=False)
        )

        self.machine.default_platform.set_pulse_on_hit_and_enable_and_release_rule.assert_called_once_with(
            SwitchSettings(hw_switch=self.machine.switches["s_left_flipper"].hw_switch, invert=False, debounce=False),
            DriverSettings(hw_driver=self.machine.coils["c_flipper_left_hold"].hw_driver,
                           pulse_settings=PulseSettings(power=1.0, duration=10),
                           hold_settings=HoldSettings(power=1.0), recycle=False)
        )

        self.machine.default_platform.clear_hw_rule = MagicMock()
        self.machine.flippers["left_flipper"].disable()
        self.assertFalse(self.machine.flippers["left_flipper"]._enabled)

        self.machine.default_platform.clear_hw_rule.assert_has_calls(
            [call(
                SwitchSettings(hw_switch=self.machine.switches["s_left_flipper"].hw_switch, invert=False, debounce=False),
                DriverSettings(hw_driver=self.machine.coils["c_flipper_left_main"].hw_driver,
                               pulse_settings=PulseSettings(power=1.0, duration=30),
                               hold_settings=None, recycle=False)
            ),
             call(
                 SwitchSettings(hw_switch=self.machine.switches["s_left_flipper"].hw_switch, invert=False, debounce=False),
                 DriverSettings(hw_driver=self.machine.coils["c_flipper_left_hold"].hw_driver,
                                pulse_settings=PulseSettings(power=1.0, duration=10),
                                hold_settings=HoldSettings(power=1.0), recycle=False)
            )
        ], any_order=True)

        self.machine.default_platform.set_pulse_on_hit_and_release_rule = MagicMock()
        self.machine.default_platform.set_pulse_on_hit_and_enable_and_release_rule = MagicMock()

        self.machine.flippers["left_flipper"].enable()
        self.machine.default_platform.set_pulse_on_hit_and_release_rule.assert_called_once_with(
            SwitchSettings(hw_switch=self.machine.switches["s_left_flipper"].hw_switch, invert=False, debounce=False),
            DriverSettings(hw_driver=self.machine.coils["c_flipper_left_main"].hw_driver,
                           pulse_settings=PulseSettings(power=1.0, duration=30),
                           hold_settings=None, recycle=False)
        )

        self.machine.default_platform.set_pulse_on_hit_and_enable_and_release_rule.assert_called_once_with(
            SwitchSettings(hw_switch=self.machine.switches["s_left_flipper"].hw_switch, invert=False, debounce=False),
            DriverSettings(hw_driver=self.machine.coils["c_flipper_left_hold"].hw_driver,
                           pulse_settings=PulseSettings(power=1.0, duration=10),
                           hold_settings=HoldSettings(power=1.0), recycle=False)
        )



from unittest.mock import MagicMock
from mpf.tests.MpfTestCase import MpfTestCase
import mpf.platforms.pololu_maestro


class TestPololuMaestro(MpfTestCase):

    def get_config_file(self):
        return 'pololu_maestro.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/pololu_maestro/'

    def get_platform(self):
        return False

    def setUp(self):
        self.serial = MagicMock()
        mpf.platforms.pololu_maestro.serial = MagicMock()
        mpf.platforms.pololu_maestro.serial.Serial.return_value = self.serial
        super().setUp()

    def _build_message(self, command, number, value, controller=12):
        lsb = value & 0x7f  # 7 bits for least significant byte
        msb = (value >> 7) & 0x7f  # shift 7 and take next 7 bits for msb
        # Send Pololu intro, device number, command, channel, and target
        # lsb/msb
        return bytes([0xaa, controller, command, number, lsb, msb])

    def test_servo_go_to_position(self):
        # go to position 1.0 (on of the ends)
        self.machine.servos["servo1"].go_to_position(1.0)
        # assert that platform got called
        self.serial.write.assert_called_with(self._build_message(0x04, 1, 9000))
        # go to position 0.0 (other end)
        self.machine.servos["servo1"].go_to_position(0.0)
        # assert that platform got called
        self.serial.write.assert_called_with(self._build_message(0x04, 1, 3000))

        self.serial.reset_mock()
        # go to position 1.0 (on of the ends)
        self.machine.servos["servo2"].go_to_position(1.0)
        # assert that platform got called
        self.serial.write.assert_called_with(self._build_message(0x04, 2, 7800))
        # go to position 0.0 (other end)
        self.machine.servos["servo2"].go_to_position(0.0)
        # assert that platform got called
        self.serial.write.assert_called_with(self._build_message(0x04, 2, 4200))
        # go to position 0.0 (middle)
        self.machine.servos["servo2"].go_to_position(0.5)
        # assert that platform got called
        self.serial.write.assert_called_with(self._build_message(0x04, 2, 6000))

    def test_daisy_chaining(self):
        # go to position 1.0 (on of the ends)
        self.machine.servos["servo1_controller_13"].go_to_position(1.0)
        # assert that platform got called
        self.serial.write.assert_called_with(self._build_message(0x04, 1, 9000, controller=13))

    def test_servo_set_speed(self):
        # test setting speed in config
        self.assertEqual(0.5, self.machine.servos["servo1"].speed_limit)
        # test standard value
        self.assertEqual(-1.0, self.machine.servos["servo2"].speed_limit)

        self.machine.servos["servo1"].set_speed_limit(-1.0)
        self.serial.write.assert_called_with(self._build_message(0x07, 1, 0))
        self.machine.servos["servo1"].set_speed_limit(0.0)
        self.serial.write.assert_called_with(self._build_message(0x07, 1, 1))
        self.machine.servos["servo1"].set_speed_limit(0.5)
        self.serial.write.assert_called_with(self._build_message(0x07, 1, 18))
        self.machine.servos["servo1"].set_speed_limit(1.0)

    def test_servo_set_acceleration(self):
        # test setting acceleration in config
        self.assertEqual(0.5, self.machine.servos["servo1"].acceleration_limit)
        # test standard value
        self.assertEqual(-1.0, self.machine.servos["servo2"].acceleration_limit)

        self.machine.servos["servo1"].set_speed_limit(-1.0)
        self.serial.write.assert_called_with(self._build_message(0x07, 1, 0))
        self.machine.servos["servo1"].set_speed_limit(0.0)
        self.serial.write.assert_called_with(self._build_message(0x07, 1, 1))
        self.machine.servos["servo1"].set_acceleration_limit(0.5)
        self.serial.write.assert_called_with(self._build_message(0x09, 1, 180))
        self.machine.servos["servo1"].set_acceleration_limit(1.0)
        self.serial.write.assert_called_with(self._build_message(0x09, 1, 255))



"""MPF plugin which automatically plays back switch events from the config file."""

import logging
from mpf.core.delays import DelayManager
from mpf.core.utility_functions import Util


class SwitchPlayer:

    """Plays back switch sequences from a config file, used for testing."""

    def __init__(self, machine):
        """Initialise switch player."""
        self.log = logging.getLogger('switch_player')

        if 'switch_player' not in machine.config:
            machine.log.debug('"switch_player:" section not found in '
                              'machine configuration, so the Switch Player'
                              'plugin will not be used.')
            return

        self.machine = machine
        self.delay = DelayManager(self.machine)
        self.current_step = 0

        self.config = self.machine.config['switch_player']
        self.machine.config_validator.validate_config("switch_player", self.config)

        self.machine.events.add_handler(self.config['start_event'],
                                        self._start_event_callback)

        self.step_list = self.config['steps']

    def __repr__(self):
        """Return string representation."""
        return '<SwitchPlayer>'

    def _start_event_callback(self, **kwargs):
        del kwargs
        self.delay.add(name='switch_player_next_step',
                       ms=Util.string_to_ms(self.step_list[self.current_step]['time']),
                       callback=self._do_step)

    def _do_step(self):

        this_step = self.step_list[self.current_step]

        self.log.debug("Switch: %s, Action: %s", this_step['switch'],
                       this_step['action'])

        # send this step's switches
        if this_step['action'] == 'activate':
            self.machine.switch_controller.process_switch(
                this_step['switch'],
                state=1,
                logical=True)
        elif this_step['action'] == 'deactivate':
            self.machine.switch_controller.process_switch(
                this_step['switch'],
                state=0,
                logical=True)
        elif this_step['action'] == 'hit':
            self._hit(this_step['switch'])

        # inc counter
        if self.current_step < len(self.step_list) - 1:
            self.current_step += 1

            # schedule next step
            self.delay.add(name='switch_player_next_step',
                           ms=Util.string_to_ms(self.step_list[self.current_step]['time']),
                           callback=self._do_step)

    def _hit(self, switch):
        self.machine.switch_controller.process_switch(
            switch,
            state=1,
            logical=True)
        self.machine.switch_controller.process_switch(
            switch,
            state=0,
            logical=True)



"""Pololu Maestro servo controller platform."""
import math
import logging
import serial
from mpf.platforms.interfaces.servo_platform_interface import ServoPlatformInterface

from mpf.core.platform import ServoPlatform


class PololuMaestroHardwarePlatform(ServoPlatform):

    """Supports the Pololu Maestro servo controllers via PySerial.

    Works with Micro Maestro 6, and Mini Maestro 12, 18, and 24.
    """

    __slots__ = ["config", "platform", "serial"]

    def __init__(self, machine):
        """Initialise Pololu Servo Controller platform."""
        super().__init__(machine)
        self.config = self.machine.config_validator.validate_config("pololu_maestro",
                                                                    self.machine.config['pololu_maestro'])
        self.platform = None
        self.serial = None
        self.features['tickless'] = True
        self._configure_device_logging_and_debug("Pololu Maestro", self.config)

    def __repr__(self):
        """Return string representation."""
        return '<Platform.Pololu_Maestro>'

    async def initialize(self):
        """Initialise platform."""
        await super().initialize()
        self.serial = serial.Serial(self.config['port'])

    def stop(self):
        """Close serial."""
        if self.serial:
            self.serial.close()
            self.serial = None

    async def configure_servo(self, number: str):
        """Configure a servo device in platform.

        Args:
        ----
            number: Number of the servo.
        """
        try:
            controller_str, number_str = number.split("-")
        except ValueError:
            number_str = number
            controller_str = "12"

        return PololuServo(int(controller_str), int(number_str), self.config, self.serial)


class PololuServo(ServoPlatformInterface):

    """A servo on the pololu servo controller."""

    __slots__ = ["log", "number", "controller_number", "config", "serial", "cmd_header"]

    def __init__(self, controller_number, number, config, serial_port):
        """Initialise Pololu servo."""
        self.log = logging.getLogger('PololuServo')
        self.number = number
        self.controller_number = controller_number
        self.config = config
        self.serial = serial_port
        self.cmd_header = bytes([0xaa, self.controller_number])

    def go_to_position(self, position):
        """Set channel to a specified target value.

        Servo will begin moving
        based on Speed and Acceleration parameters previously set.
        Target values will be constrained within Min and Max range, if set.
        For servos, target represents the pulse width in of
        quarter-microseconds.
        Servo center is at 1500 microseconds, or 6000 quarter-microseconds
        Typcially valid servo range is 3000 to 9000 quarter-microseconds
        If channel is configured for digital output, values < 6000 = Low ouputco.

        Args:
        ----
            position: Servo position between 0 and 1
        """
        servo_min = self.config['servo_min']
        servo_max = self.config['servo_max']
        value = int(servo_min + position * (servo_max - servo_min))

        # if Min is defined and Target is below, force to Min
        if (self.config['servo_min'] > 0 and
                value < self.config['servo_min']):
            value = self.config['servo_min']

        # if Max is defined and Target is above, force to Max
        if 0 < self.config['servo_max'] < value:
            value = self.config['servo_max']

        lsb = value & 0x7f  # 7 bits for least significant byte
        msb = (value >> 7) & 0x7f  # shift 7 and take next 7 bits for msb
        # Send Pololu intro, device number, command, channel, and target
        # lsb/msb
        cmd = self.cmd_header + bytes([0x04, self.number, lsb, msb])
        if self.config['debug']:
            self.log.debug("Sending cmd: %s", "".join(" 0x%02x" % b for b in cmd))
        self.serial.write(cmd)

    def set_speed_limit(self, speed_limit):
        """Set the speed of the channel.

        Speed is measured as 0.25microseconds/10milliseconds

        For the standard 1ms pulse width change to move a servo between
        extremes, a speed of 1 will take 1 minute, and a speed of 60 would take
        1 second.

        Speed of 0 is unrestricted.

        Args:
        ----
            speed_limit: speed_limit to set

        """
        if speed_limit == -1.0:
            maestro_speed_limit = 0  # 0 is unrestricted for the maestro
        elif speed_limit == 0:
            maestro_speed_limit = 1  # minimum speed setting
        else:
            max_pos_change_per_second = speed_limit * self.config['servo_max']  # change normalized values for maestro
            maestro_speed_limit = int(max_pos_change_per_second / 1000 / 0.25)

        lsb = maestro_speed_limit & 0x7f  # 7 bits for least significant byte
        msb = (maestro_speed_limit >> 7) & 0x7f  # shift 7 and take next 7 bits for msb
        cmd = self.cmd_header + bytes([0x07, self.number, lsb, msb])
        self.serial.write(cmd)

    def set_acceleration_limit(self, acceleration_limit):
        """Set acceleration of channel.

        This provide soft starts and finishes when servo moves to target
        position.

        Valid values are from 0 to 255. 0=unrestricted, 1 is slowest start.
        It is measured in units of 0.25microseconds/10milliseconds/80milliseconds
        A value of 1 will take the servo about 3s to move between 1ms to 2ms
        range.
        """
        if acceleration_limit == -1.0:
            maestro_acceleration_normalized = 0  # 0 is unrestricted for the maestro
        elif acceleration_limit == 0:
            maestro_acceleration_normalized = 1  # minimum acceleration setting
        else:
            max_speed_change_per_second = math.sqrt(acceleration_limit * self.config['servo_max'])
            maestro_acceleration_limit = calculate_maestro_acceleration(max_speed_change_per_second)
            max_limit_value = calculate_maestro_acceleration(math.sqrt(1.0 * self.config['servo_max']))
            maestro_acceleration_normalized = int(255 / max_limit_value * maestro_acceleration_limit)

        lsb = maestro_acceleration_normalized & 0x7f  # 7 bits for least significant byte
        msb = (maestro_acceleration_normalized >> 7) & 0x7f  # shift 7 and take next 7 bits for msb
        cmd = self.cmd_header + bytes([0x09, self.number, lsb, msb])
        self.serial.write(cmd)


def calculate_maestro_acceleration(normalized_limit):
    """Calculate acceleration limit for the maestro."""
    return normalized_limit / 1000 / 0.25 / 80



"""Test assets."""
import time
from mpf.tests.MpfTestCase import MpfTestCase


class TestAssets(MpfTestCase):
    def get_machine_path(self):
        return 'tests/machine_files/asset_manager'

    def get_config_file(self):
        return 'test_asset_loading.yaml'

    def test_asset_loading(self):
        # TODO: instantiate a fresh machine between test groups
        self.expected_duration = 1.5
        self._test_machine_wide_asset_loading()
        self._test_random_asset_group()
        self._test_random_asset_group_with_weighting()
        self._test_random_force_all()
        self._test_random_force_next()
        self._test_sequence_asset_group()
        self._test_sequence_asset_group_with_count()
        self._test_conditional_random_asset_group()
        self._test_conditional_sequence_asset_group()

    def _test_machine_wide_asset_loading(self):

        # test that the shows asset class gets built correctly
        self.assertTrue(self.machine, 'shows')
        # tests that assets are registered as expected with various conditions

        # /shows folder
        self.assertIn('show1', self.machine.shows)
        self.assertIn('show2', self.machine.shows)
        self.assertIn('show3', self.machine.shows)

        # test subfolders listed in assets:shows machine-wide config folders
        self.assertIn('show4', self.machine.shows)  # /shows/preload
        self.assertIn('show4b', self.machine.shows)  # /shows/preload/subfolder
        self.assertIn('show5', self.machine.shows)  # /shows/on_demand

        # test shows from subfolder not listed in assets:shows
        self.assertIn('show11', self.machine.shows)  # /shows/custom1

        self.assertIn('show12', self.machine.shows)
        self.assertIn('show13', self.machine.shows)

        # Test that mode assets were loaded properly
        self.assertIn('show6', self.machine.shows)
        self.assertIn('show7', self.machine.shows)
        self.assertIn('show8', self.machine.shows)
        self.assertIn('show9', self.machine.shows)
        self.assertIn('show10', self.machine.shows)

    def _test_random_asset_group(self):
        # three assets, no weights

        # make sure the asset group was created
        self.assertIn('group1', self.machine.shows)

        # make sure the randomness is working. To test this, we request the
        # asset 10,000 times and then count the results and assume that each
        # should be 3,333 +- 500 just to make sure the test never fails/
        res = list()
        for x in range(10000):
            res.append(self.machine.shows['group1'].asset)

        self.assertAlmostEqual(3333, res.count(self.machine.shows['show1']),
                               delta=500)
        self.assertAlmostEqual(3333, res.count(self.machine.shows['show2']),
                               delta=500)
        self.assertAlmostEqual(3333, res.count(self.machine.shows['show3']),
                               delta=500)

    def _test_random_asset_group_with_weighting(self):
        # three assets, third one has a weight of 2

        # make sure the asset group was created
        self.assertIn('group2', self.machine.shows)

        # make sure the randomness is working. To test this, we request the
        # asset 10,000 times and then count the results and assume that each
        # should be 3,333 +- 500 just to make sure the test never fails/
        res = list()
        for x in range(10000):
            res.append(self.machine.shows['group2'].asset)

        self.assertAlmostEqual(2500, res.count(self.machine.shows['show1']),
                               delta=500)
        self.assertAlmostEqual(2500, res.count(self.machine.shows['show2']),
                               delta=500)
        self.assertAlmostEqual(5000, res.count(self.machine.shows['show3']),
                               delta=500)

    def _test_sequence_asset_group(self):
        # three assets, no weights

        self.assertIn('group3', self.machine.shows)

        # Should always return in order, 1, 2, 3, 1, 2, 3...
        self.assertIs(self.machine.shows['group3'].asset, self.machine.shows['show1'])
        self.assertIs(self.machine.shows['group3'].asset, self.machine.shows['show2'])
        self.assertIs(self.machine.shows['group3'].asset, self.machine.shows['show3'])
        self.assertIs(self.machine.shows['group3'].asset, self.machine.shows['show1'])
        self.assertIs(self.machine.shows['group3'].asset, self.machine.shows['show2'])
        self.assertIs(self.machine.shows['group3'].asset, self.machine.shows['show3'])
        self.assertIs(self.machine.shows['group3'].asset, self.machine.shows['show1'])
        self.assertIs(self.machine.shows['group3'].asset, self.machine.shows['show2'])
        self.assertIs(self.machine.shows['group3'].asset, self.machine.shows['show3'])

    def _test_sequence_asset_group_with_count(self):
        # three assets, no weights

        self.assertIn('group4', self.machine.shows)

        # Should always return in order, 1, 1, 1, 1, 2, 2, 3, 1, 1, 1, 1 ...
        self.assertIs(self.machine.shows['group4'].asset, self.machine.shows['show1'])
        self.assertIs(self.machine.shows['group4'].asset, self.machine.shows['show1'])
        self.assertIs(self.machine.shows['group4'].asset, self.machine.shows['show1'])
        self.assertIs(self.machine.shows['group4'].asset, self.machine.shows['show1'])
        self.assertIs(self.machine.shows['group4'].asset, self.machine.shows['show2'])
        self.assertIs(self.machine.shows['group4'].asset, self.machine.shows['show2'])
        self.assertIs(self.machine.shows['group4'].asset, self.machine.shows['show3'])
        self.assertIs(self.machine.shows['group4'].asset, self.machine.shows['show1'])
        self.assertIs(self.machine.shows['group4'].asset, self.machine.shows['show1'])
        self.assertIs(self.machine.shows['group4'].asset, self.machine.shows['show1'])
        self.assertIs(self.machine.shows['group4'].asset, self.machine.shows['show1'])
        self.assertIs(self.machine.shows['group4'].asset, self.machine.shows['show2'])
        self.assertIs(self.machine.shows['group4'].asset, self.machine.shows['show2'])
        self.assertIs(self.machine.shows['group4'].asset, self.machine.shows['show3'])

    def _test_random_force_next(self):
        # random, except it ensures the same one does not show up twice in a
        # row

        self.assertIn('group5', self.machine.shows)

        # do it 10,000 times just to be sure. :)
        last = self.machine.shows['group5'].asset
        res = list()

        for x in range(10000):
            show = self.machine.shows['group5'].asset
            self.assertIsNot(last, show)
            last = show

            res.append(show)

        # Also check that the weights were right

        # BTW these weights are non-intuitive since the last asset is not
        # considered for the next round. e.g. show1 = 1, show2 = 5,
        # show3 = 1, so you'd think they would be 1400, 7200, 1400, but in
        # reality, 50% of the time, asset2 is not in contention, so really
        # asset2 has a 6-to-1 (84%) chance of being selected 66% of the time,
        # but a 0% chance of being selected 33% of the time, etc. So trust that
        # these numbers are right. :)
        self.assertAlmostEqual(2733, res.count(self.machine.shows['show1']),
                               delta=500)
        self.assertAlmostEqual(4533, res.count(self.machine.shows['show2']),
                               delta=500)
        self.assertAlmostEqual(2733, res.count(self.machine.shows['show3']),
                               delta=500)

    def _test_random_force_all(self):
        # random, except it ensures the same one does not show up twice before
        # they're all shown

        self.assertIn('group6', self.machine.shows)

        for x in range(1000):
            this_set = set()
            this_set.add(self.machine.shows['group6'].asset)
            this_set.add(self.machine.shows['group6'].asset)
            this_set.add(self.machine.shows['group6'].asset)

            self.assertEqual(len(this_set), 3)

    def _test_conditional_random_asset_group(self):

        # make sure the asset group was created
        self.assertIn('group1', self.machine.shows)

        # ONE valid show
        # Request the show 1,000 times and ensure that only one show was picked
        res = list()
        for x in range(1000):
            res.append(self.machine.shows['group7'].asset)

        self.assertEqual(1000, res.count(self.machine.shows['show1']))
        self.assertEqual(0, res.count(self.machine.shows['show2']))
        self.assertEqual(0, res.count(self.machine.shows['show3']))

        # TWO valid shows
        # Request the show 10,000 times and ensure that two shows are fairly split
        self.machine.modes["mode1"].start()
        self.advance_time_and_run()
        res = list()
        for x in range(10000):
            res.append(self.machine.shows['group7'].asset)

        self.assertAlmostEqual(5000, res.count(self.machine.shows['show1']),
                               delta=250)
        self.assertAlmostEqual(5000, res.count(self.machine.shows['show2']),
                               delta=250)
        self.assertEqual(0, res.count(self.machine.shows['show3']))

        # THREE valid shows
        # Request the show 10,000 times and ensure that all three shows are fairly split
        self.machine.modes["mode1"].stop()
        res = list()
        for x in range(10000):
            res.append(self.machine.shows['group7'].asset)

        self.assertAlmostEqual(3333, res.count(self.machine.shows['show1']),
                               delta=250)
        self.assertAlmostEqual(3333, res.count(self.machine.shows['show2']),
                               delta=250)
        self.assertAlmostEqual(3333, res.count(self.machine.shows['show3']),
                               delta=250)

        # play a group
        self.machine.shows['group7'].play()

    def _test_conditional_sequence_asset_group(self):
        # These tests are not independent, and mode1 is still running/stopping from the above test :(
        self.advance_time_and_run()
        self.assertIn('group8', self.machine.shows)

        # ONE valid show
        self.assertIs(self.machine.shows['group8'].asset, self.machine.shows['show2'])
        self.assertIs(self.machine.shows['group8'].asset, self.machine.shows['show2'])
        self.assertIs(self.machine.shows['group8'].asset, self.machine.shows['show2'])
        self.assertIs(self.machine.shows['group8'].asset, self.machine.shows['show2'])
        self.assertIs(self.machine.shows['group8'].asset, self.machine.shows['show2'])
        self.assertIs(self.machine.shows['group8'].asset, self.machine.shows['show2'])
        self.assertIs(self.machine.shows['group8'].asset, self.machine.shows['show2'])
        self.assertIs(self.machine.shows['group8'].asset, self.machine.shows['show2'])
        self.assertIs(self.machine.shows['group8'].asset, self.machine.shows['show2'])

        # TWO valid shows
        self.machine.modes["mode1"].start()
        self.advance_time_and_run()
        self.assertIs(self.machine.shows['group8'].asset, self.machine.shows['show1'])
        self.assertIs(self.machine.shows['group8'].asset, self.machine.shows['show2'])
        self.assertIs(self.machine.shows['group8'].asset, self.machine.shows['show1'])
        self.assertIs(self.machine.shows['group8'].asset, self.machine.shows['show2'])
        self.assertIs(self.machine.shows['group8'].asset, self.machine.shows['show1'])
        self.assertIs(self.machine.shows['group8'].asset, self.machine.shows['show2'])
        self.assertIs(self.machine.shows['group8'].asset, self.machine.shows['show1'])
        self.assertIs(self.machine.shows['group8'].asset, self.machine.shows['show2'])

        # THREE valid shows
        self.machine.modes["mode1"].stop()
        self.assertIs(self.machine.shows['group8'].asset, self.machine.shows['show3'])
        self.assertIs(self.machine.shows['group8'].asset, self.machine.shows['show1'])
        self.assertIs(self.machine.shows['group8'].asset, self.machine.shows['show2'])
        self.assertIs(self.machine.shows['group8'].asset, self.machine.shows['show3'])
        self.assertIs(self.machine.shows['group8'].asset, self.machine.shows['show1'])
        self.assertIs(self.machine.shows['group8'].asset, self.machine.shows['show2'])
        self.assertIs(self.machine.shows['group8'].asset, self.machine.shows['show3'])
        self.assertIs(self.machine.shows['group8'].asset, self.machine.shows['show1'])
        self.assertIs(self.machine.shows['group8'].asset, self.machine.shows['show2'])
        self.assertIs(self.machine.shows['group8'].asset, self.machine.shows['show3'])

